"""
karta.raster

Classes for grids of raster data
"""

import numpy
import numbers
numbers.Integral.register(numpy.integer)

from . import grid
from . import misc

from .grid import RegularGrid, merge, gridpoints, mask_poly
from .band import SimpleBand, CompressedBand
from .read import read_aai, read_geotiff, read_gtiff, from_geotiffs
from .misc import (normed_potential_vectors,
                   slope, aspect, gradient, divergence, hillshade)

__all__ = ["grid", "misc", "RegularGrid",
           "read_aai", "read_geotiff", "read_gtiff", "from_geotiffs",
           "slope", "aspect", "gradient", "divergence", "hillshade",
           "normed_potential_vectors"]





def _flatten(vertices):
    """ Convert a nested list of coordinates into a flat list. """
    out = []
    for item in vertices:
        if hasattr(item[0], "__iter__"):
            verts = _flatten(item)
            out.extend(verts)
        else:
            out.append(item)
    return out

def _as_nested_lists(vertices):
    """ Convert a nested structure such as an ndarray into a list of lists. """
    out = []
    for part in vertices:
        if hasattr(part[0], "__iter__"):
            verts = _as_nested_lists(part)
            out.append(verts)
        else:
            out.append(list(part))
    return out

def _reproject(xy, crs1, crs2):
    """ Reproject a coordinate (or 2-tuple of x and y vectors) from *crs1* to
    *crs2*. """
    return crs1.transform(crs2, *xy)

def _reproject_nested(xy, crs1, crs2):
    """ Reproject a possibly nested list of coordinates, retaining the nesting
    structure. """
    if hasattr(xy[0], "__iter__"):
        out = []
        for xy_ in xy:
            out.append(_reproject_nested(xy_, crs1, crs2))
        return out
    else:
        return _reproject(xy, crs1, crs2)




"""
Raster grid representations
"""
import copy
import math
import numbers
import warnings
import numpy as np
from . import _gdal
from . import crfuncs
from .band import SimpleBand, CompressedBand, BandIndexer
from .coordgen import CoordinateGenerator
from .. import errors
from ..crs import Cartesian

BAND_CLASS_DEFAULT = CompressedBand
CRS_DEFAULT = Cartesian

class Grid(object):
    """ Grid base class """

    @property
    def nodata(self):
        return self._nodata

    def max(self):
        """ Return the maximum non-nan in self.data """
        tmp = self[self.data_mask]
        if len(tmp) != 0:
            return tmp.max()
        else:
            return np.nan

    def min(self):
        """ Return the minimum non-nan in self.data """
        tmp = self[self.data_mask]
        if len(tmp) != 0:
            return tmp.min()
        else:
            return np.nan

    def minmax(self):
        """ Return the minimum and maximum value of data array """
        tmp = self[self.data_mask_full]
        if len(tmp) != 0:
            return (tmp.min(), tmp.max())
        else:
            return (np.nan, np.nan)

    def copy(self):
        """ Return a deep copy """
        return copy.deepcopy(self)

    def apply(self, func, inplace=False):
        """ Apply a vector function to grid values.

        Parameters
        ----------
        func : callable
            function that takes a vector of values and returns an equally-sized
            vector of values with the same type
        inplace : boolean, optional
            whether to perform operation in place rather than returning a new
            grid instance (default False)

        Returns
        -------
        instance of type(self)
        """
        msk = self.data_mask_full

        if inplace:
            self[msk] = func(self[msk])
            return self
        else:
            newgrid = self.copy()
            newgrid[msk] = func(self[msk])
            return newgrid


class RegularGrid(Grid):
    """ Regular  structured grid class. A RegularGrid contains a fixed number
    of rows and columns with a constant spacing and a list of bands
    representing scalar fields.

    Positions on a RegularGrid are referenced to their array indices (i,j)
    using an affine transform *T* such that

    ::

        T = (a, b, c, d, e, f)

        x = a + j*c + i*e
        y = b + i*d + j*f

    or as a matrix transformation,

    ::

        |e  c  a|   |i|   |x|
        |       | * |j| = | |
        |d  f  b|   |1|   |y|

    where (x,y) are the coordinates of the lower left corner of each pixel.

    Meaning:

    - (a,b) defines the lower-left grid cell corner
    - (c,d) defines the resolution in the horizontal and vertical directions
    - (e,f) can be used to define a rotation

    In the common case of a "east-right, north-up" grid, e = f = 0.
    """
    def __init__(self, transform, values=None, bands=None, crs=None,
            nodata_value=None, bandclass=None):
        """ Create a RegularGrid instance.

        Parameters
        ----------
        transform : 6-tuple of floats
            a six element list, tuple, or dictionary defining the geotransform
            of the grid: ``[xllcorner, yllcorner, dx, dy, sx, sy]``
        values : ndarray (nrows x ncols), optional
            grid data from which raster bands will be formed. A two-dimensional
            array will be transformed into a single band, while a
            three-dimension MxNxP array will become P MxN bands. The structure
            used to store band data internally can be controlled by the
            *bandclass* keyword argument.
        bands : list of band types, optional
            list of allocated Band-like objects. If provided, they are assumed
            to be of a single type, which overrides *bandclass*.
        crs : karta.crs.CRS subclass, optional
            Grid coordinate reference system. default CRS_DEFAULT
        nodata_value : number, optional
            specifies a value to represent NoData or null pixels. The default
            is chosen depending on the input type of *values* or *bands*. If
            neither is provided, the default is NaN.
        bandclass : class, optional
            indicates the band class used to represent grid data. default
            BAND_CLASS_DEFAULT
        """

        if hasattr(transform, "keys"):
            self._transform = tuple([float(transform[f]) for f in
                    ("xllcorner", "yllcorner", "dx", "dy", "sx", "sy")])
        elif len(transform) == 6:
            self._transform = tuple(float(a) for a in transform)
        else:
            raise errors.GridError("RegularGrid must be initialized with a "
                                   "transform iterable or dictionary")

        if bands is not None:
            self._bndcls = type(bands[0])
        elif bandclass is None:
            self._bndcls = BAND_CLASS_DEFAULT
        else:
            self._bndcls = bandclass

        if bands is not None:
            if any(bands[0].size != b.size for b in bands[1:]):
                size_str = " ".join(set(str(b.size) for b in bands))
                raise ValueError("bands have mismatching size: {0}".format(size_str))
            self.bands = bands
        else:
            self.bands = []

        if bands is None and (values is not None):
            if values.ndim == 2:
                band = self._bndcls(values.shape, values.dtype.type)
                band.setblock(0, 0, values)
                self.bands.append(band)
            elif values.ndim == 3:
                for iband in range(values.shape[2]):
                    band = self._bndcls(values.shape[:2], values.dtype.type)
                    band.setblock(0, 0, values[:,:,iband])
                    self.bands.append(band)
            else:
                raise ValueError("`values` must have two or three dimensions")

        self._bandindexer = BandIndexer(self.bands)

        if crs is None:
            self.crs = CRS_DEFAULT
        else:
            self.crs = crs

        if nodata_value is None:
            if len(self.bands) == 0:
                self._nodata = np.nan
            else:
                self._nodata = get_nodata(self.bands[0].dtype)
        else:
            self._nodata = nodata_value
        return

    def __add__(self, other):
        if self._equivalent_structure(other):
            return RegularGrid(copy.copy(self.transform),
                               values=self[:,:]+other[:,:],
                               crs=self.crs, nodata_value=self.nodata)
        else:
            raise ValueError(self, other)

    def __sub__(self, other):
        if self._equivalent_structure(other):
            return RegularGrid(copy.copy(self.transform),
                               values=self[:,:]-other[:,:],
                               crs=self.crs, nodata_value=self.nodata)
        else:
            raise ValueError(self, other)

    def __getitem__(self, key):
        return self._bandindexer[key]

    def __setitem__(self, key, value):
        self._bandindexer[key] = value
        return

    def _equivalent_structure(self, other):
        return (self._transform == other._transform) and \
               (self.size == other.size)

    @property
    def transform(self):
        return self._transform

    @property
    def values(self):
        return self._bandindexer

    @property
    def size(self):
        return self.bands[0].size

    @property
    def nbands(self):
        return len(self.bands)

    @property
    def nodata_value(self):
        return self._nodata

    def set_nodata_value(self, val):
        """ Redefine value used to indicate nodata.

        Parameters
        ----------
        val : number
        """
        self[:,:,:] = np.where(self.data_mask_full, self[:,:,:], val)
        self._nodata = val
        return

    def center_llref(self):
        """ Return the 'lower-left' reference in terms of a center coordinate.
        """
        T = self._transform
        xllcenter = self._transform[0] + 0.5 * (T[2] + T[4])
        yllcenter = self._transform[1] + 0.5 * (T[3] + T[5])
        return xllcenter, yllcenter

    def corner_llref(self):
        """ Return the 'lower-left' reference in terms of a center coordinate.
        """
        return self._transform[:2]

    def coordmesh(self, *args, **kwargs):
        """ Alias for center_coords() """
        return self.center_coords(*args, **kwargs)

    def coordinates(self, crs=None):
        """ Return a CoordinateGenerator instance for computing the coordinates
        of indices in *crs*. This behaves like the array returned by
        `center_coords()`, but computes coordinates on demand rather than at
        once.

        Parameters
        ----------
        crs : CRS, optional
            Coordinate system of output coordinates. If None (default), output
            coordinates are not transformed from *self.crs*.

        Returns
        -------
        CoordinateGenerator

        Notes
        -----
        This is an experimental replacement for coordmesh() and center_coords().
        """
        if crs is None:
            crs = self.crs
        return CoordinateGenerator(self.transform, self.size, self.crs, crs)

    def center_coords(self):
        """ Return the coordinates of cell centers. """
        t = self._transform
        xcoords = np.empty(self.size)
        ycoords = np.empty(self.size)
        icol = np.arange(self.size[1])
        for i in range(self.size[0]):
            xcoords[i,:] = t[0] + (icol+0.5)*t[2] + (i+0.5)*t[4]
            ycoords[i,:] = (t[1] + (i+0.5)*t[3] + (icol+0.5)*t[5])
        return xcoords, ycoords

    def vertex_coords(self):
        """ Return the coordinates of cell vertices. """
        t = self._transform
        ny, nx = self.size
        xcoords = np.empty((ny+1, nx+1))
        ycoords = np.empty((ny+1, nx+1))
        icol = np.arange(0, self.size[1]+1)
        for i in range(self.size[0]):
            xcoords[i,:] = t[0] + icol*t[2] + i*t[4]
            ycoords[i,:] = (t[1] + i*t[3] + icol*t[5])
        return xcoords, ycoords

    @property
    def origin(self):
        return self.transform[:2]

    @property
    def resolution(self):
        return self.transform[2:4]

    @property
    def skew(self):
        return self.transform[4:]

    @property
    def bbox(self):
        extent = self.get_extent(reference="edge")
        return extent[0], extent[2], extent[1], extent[3]

    @property
    def data_bbox(self):
        extent = self.get_data_extent(reference="edge")
        return extent[0], extent[2], extent[1], extent[3]

    @property
    def extent(self):
        return self.get_extent()

    def get_bbox(self, crs=None):
        a, b, c, d = self.get_extent(reference="edge", crs=crs)
        return a, c, b, d

    def get_extent(self, reference='center', crs=None):
        """ Return the region characteristics as a tuple (xmin, xmax, ymin,
        ymax).

        Parameters
        ----------
        reference : string
            either 'center' or 'edge'.
        crs : karta.crs.CRS subclass
            coordinate system of output
        """
        if reference == 'center':
            x0, y0 = self.center_llref()
            n = -1
        elif reference == 'edge':
            x0, y0 = self.corner_llref()
            n = 0
        else:
            raise errors.GridError("`reference` must be 'center' or 'edge'")
        ny, nx = self.size
        dx, dy = self._transform[2:4]
        sx, sy = self._transform[4:]

        sgn = lambda a: 0 if a==0 else a/abs(a)
        if sgn(dx) == sgn(sx):
            x1 = x0 + dx*(nx+n) + sx*(ny+n)
        else:
            x1 = x0 + dx*(nx+n) - sx*(ny+n)

        if sgn(dy) == sgn(sy):
            y1 = y0 + dy*(ny+n) + sy*(nx+n)
        else:
            y1 = y0 + dy*(ny+n) - sy*(nx+n)

        if crs is None or (crs == self.crs):
            (a, b, c, d) = min(x0, x1), max(x0, x1), min(y0, y1), max(y0, y1)
        else:
            xg0, yg0 = self.crs.transform(crs, x0, y0)
            xg1, yg1 = self.crs.transform(crs, x0, y1)
            xg2, yg2 = self.crs.transform(crs, x1, y0)
            xg3, yg3 = self.crs.transform(crs, x1, y1)
            a = min(xg0, xg1, xg2, xg3)
            b = max(xg0, xg1, xg2, xg3)
            c = min(yg0, yg1, yg2, yg3)
            d = max(yg0, yg1, yg2, yg3)
        return a, b, c, d

    def get_data_extent(self, reference='center', nodata=None, crs=None):
        """ Return the region characteristics as a tuple (xmin, xmax, ymin,
        ymax).

        Parameters
        ----------
        reference : string
            either 'center' or 'edge'.
        crs : karta.crs.CRS subclass
            coordinate system of output
        """
        if nodata is None:
            nodata = self.nodata

        if np.isnan(nodata):
            isdata = lambda a: ~np.isnan(a)
        else:
            def isdata(a):
                return a != nodata

        dx, dy = self.transform[2:4]
        sx, sy = self.transform[4:6]
        x0 = self.transform[0] + 0.5*dx + 0.5*sx
        y0 = self.transform[1] + 0.5*dy + 0.5*sy
        ny, nx = self.size

        # Reading a row is fast, so process from bottom to top
        bx = x0
        by = y0 + ny*dy
        tx = x0
        ty = y0
        lx = x0 + nx*dx
        ly = y0
        rx = y0
        ry = y0

        jvec = np.arange(nx)

        for i, row in enumerate(self[:,:]):
            x = (x0 + jvec*dx + i*sx)[isdata(row.squeeze())]
            y = (y0 + i*dy + jvec*sy)[isdata(row.squeeze())]

            if len(x) != 0:

                xmax, xmin = x.max(), x.min()
                ymax, ymin = y.max(), y.min()
                if ymin < by:
                    by = ymin
                    bx = x[y==ymin]
                if ymax > ty:
                    ty = ymax
                    tx = x[y==ymax]
                if xmin < lx:
                    lx = xmin
                    ly = y[x==xmin]
                if xmax > rx:
                    rx = xmax
                    ry = y[x==xmax]

        if reference == 'center':
            pass
        elif reference == 'edge':
            lx -= 0.5*dx - 0.5*sx
            rx += 0.5*dx + 0.5*sx
            by -= 0.5*dy - 0.5*sy
            ty += 0.5*dy + 0.5*sy
        else:
            raise errors.GridError("`reference` must be 'center' or 'edge'")

        if (crs is not None) and (crs != self.crs):
            lx, ly = self.crs.transform(crs, lx, ly)
            rx, ry = self.crs.transform(crs, rx, ry)
            bx, by = self.crs.transform(crs, bx, by)
            tx, ty = self.crs.transform(crs, tx, ty)
        return (lx, rx, by, ty)

    @property
    def data_mask_full(self):
        """ 8-bit mask of valid data cells by band """
        if np.isnan(self.nodata):
            isdata = lambda a: ~np.isnan(a)
        else:
            isdata = lambda a: a != self.nodata
        return np.dstack([isdata(self[:,:,i]) for i in range(len(self.bands))])

    @property
    def data_mask(self):
        """ 8-bit mask of valid data cells, collapsed across bands """
        return np.all(self.data_mask_full, axis=-1)

    def aschunks(self, size=(-1, -1), overlap=(0, 0), copy=True):
        """ Generator for grid chunks. This may be useful for parallel or
        memory-controlled grid processing.

        Parameters
        ----------
        size : tuple of two integers, optional
            size of the chunks to return (default approximately one-quarter of
            each dimension)
        overlap : tuple of two integers, optional
            number of pixels of overlap (default (0, 0))
        copy : bool
            whether to force returned grids to be copies
            warning: output may be a copy regardless, depending on the band class

        Yields
        ------
        RegularGrid
            smaller grids
        """
        ny, nx = self.size
        if size == (-1, -1):
            size = (nx//4, ny//4)

        i0 = 0
        j0 = 0

        T0 = self.transform
        while 1:
            if j0 >= nx:
                j0 = 0
                i0 += size[1]-overlap[1]

            if i0 >= ny:
                break

            T = [self.transform[0] + j0*T0[2] + i0*T0[4],
                 self.transform[1] + i0*T0[3] + j0*T0[5],
                 T0[2], T0[3], T0[4], T0[5]]
            if copy:
                v = self[i0:i0+size[1], j0:j0+size[0]].copy()
            else:
                v = self[i0:i0+size[1], j0:j0+size[0]]
            yield RegularGrid(T, values=v, crs=self.crs, nodata_value=self.nodata)
            j0 += size[0]-overlap[0]

    def clip(self, xmin, xmax, ymin, ymax, crs=None):
        """ Return a clipped version of grid with cell centers constrained to a
        bounding box.

        Parameters
        ----------
        xmin : float
        xmax : float
        ymin : float
        ymax : float
        crs : karta.crs.CRS subclass, optional
        """
        if crs is not None:
            x, y = crs.transform(self.crs, [xmin, xmin, xmax, xmax],
                                           [ymin, ymax, ymin, ymax])
            xmin = min(x)
            xmax = max(x)
            ymin = min(y)
            ymax = max(y)

        # Convert to center coords
        t = self.transform

        ll = self.get_positions(xmin, ymin)
        lr = self.get_positions(xmax, ymin)
        ul = self.get_positions(xmin, ymax)
        ur = self.get_positions(xmax, ymax)

        i0 = int(np.ceil(min(ll[0], lr[0], ul[0], ur[0])))
        i1 = int(np.floor(max(ll[0], lr[0], ul[0], ur[0]))) + 1
        j0 = int(np.ceil(min(ll[1], lr[1], ul[1], ur[1])))
        j1 = int(np.floor(max(ll[1], lr[1], ul[1], ur[1]))) + 1

        values = self[i0:i1,j0:j1].copy()
        x0 = t[0] + j0*t[2] + i0*t[4]
        y0 = t[1] + i0*t[3] + j0*t[5]
        tnew = (x0, y0, t[2], t[3], t[4], t[5])
        return RegularGrid(tnew, values, crs=self.crs, nodata_value=self.nodata)

    def resize(self, bboxnew):
        """ Return a new grid grid with outer edges given by a bounding box.

        If the grid origin shifts by a non-integer factor of the grid
        resolution, nearest neighbour values are selected.

        Parameters
        ----------
        bboxnew : 4-tuple of floats
            (xmin, ymin, xmax, ymax).
        """
        # Define own rounding function so that half-values are treated consistently
        def _round(a):
            r = a%1
            if r <= 0.5:
                return a//1
            else:
                return a//1+1

        bb = self.bbox
        bbnew = list(bboxnew)
        dx, dy, sx, sy = self.transform[2:]

        # Redefine output box dimensions to be an integer factor of dx, dy
        bbnew[2] = bbnew[0] + dx*math.ceil((bbnew[2]-bbnew[0])/dx)
        bbnew[3] = bbnew[1] + dy*math.ceil((bbnew[3]-bbnew[1])/dy)

        ny, nx = self.size
        nxnew = int(_round((bbnew[2]-bbnew[0])/dx))
        nynew = int(_round((bbnew[3]-bbnew[1])/dy))
        Tnew = [bbnew[0], bbnew[1], dx, dy, sx, sy]

        # determine the indices of existing data on the new grid
        j0new = max(0,     int(_round((bb[0]-bbnew[0])/dx)))
        j1new = min(nxnew, int(_round((bb[2]-bbnew[0])/dx)))
        i0new = max(0,     int(_round((bb[1]-bbnew[1])/dy)))
        i1new = min(nynew, int(_round((bb[3]-bbnew[1])/dy)))

        # determine the indices of the data to copy on the old grid
        j0 = max(0,  int(_round((bbnew[0]-bb[0])/dx)))
        j1 = min(nx, int(_round((bbnew[2]-bb[0])/dx)))
        i0 = max(0,  int(_round((bbnew[1]-bb[1])/dy)))
        i1 = min(ny, int(_round((bbnew[3]-bb[1])/dy)))

        newbands = []
        for band in self.bands:
            newband = self._bndcls((nynew, nxnew), dtype=band.dtype,
                                   initval=self.nodata)
            newband.setblock(i0new, j0new, band.getblock(i0, j0, i1-i0, j1-j0))
            newbands.append(newband)

        gridnew = RegularGrid(Tnew, bands=newbands, crs=self.crs,
                              nodata_value=self.nodata)
        return gridnew

    def mask_by_poly(self, polys, inplace=False):
        """ Return a grid with all elements outside the bounds of a polygon
        masked as nodata.

        Parameters
        ----------
        polys : Polygon, Multipolygon or list of Polygon instances
            region(s) defining masking boundary
        inplace : bool, optional
            if True, perform masking in-place (default False)
        """
        if getattr(polys, "_geotype", "") == "Polygon":
            polys = [polys]

        def unpack_multipolygons(L):
            out = []
            for item in L:
                if getattr(item, "_geotype", "") == "Multipolygon":
                    out.extend(item[i] for i in range(len(item)))
                elif getattr(item, "_geotype", "") == "Polygon":
                    out.append(item)
                elif isinstance(item, (list, tuple)):
                    out.extend(unpack_multipolygons(a for a in item))
            return out

        polys = unpack_multipolygons(polys)

        ny, nx = self.size
        msk = np.zeros([ny, nx], dtype=np.bool)
        for poly in polys:

            x, y = poly.get_coordinate_lists(self.crs)[:2]
            if not poly.isclockwise():
                x = x[::-1]
                y = y[::-1]

            _msk = mask_poly(x, y, nx, ny, self.transform)
            msk = (msk | _msk)

        if inplace:
            for i in range(len(self.bands)):
                data = self[:,:,i]
                data[~msk] = self.nodata
                self[:,:,i] = data
            return self
        else:
            msk = np.broadcast_to(np.atleast_3d(msk), [ny, nx, self.nbands])
            return RegularGrid(self.transform,
                               values=np.where(msk, self[:,:,:], self.nodata),
                               crs=self.crs, nodata_value=self.nodata)

    def _resample_transform(self, transform, method='nearest'):
        """ Resample grid to match a new transform.

        Arguments
        ---------
        transform : iterable
        method : str, optional
            default 'nearest'

        Returns
        -------
        RegularGrid
        """
        cg = CoordinateGenerator(transform, self.size, self.crs, self.crs)
        X, Y = cg[:,:]
        if method == 'nearest':
            values = self.sample_nearest(X, Y)
        elif method == 'linear':
            values = self.sample_bilinear(X, Y)
        else:
            raise NotImplementedError('method "{0}" unavailable'.format(method))
        if values.ndim == 3:
            values = values.transpose(1, 2, 0)
        return RegularGrid(transform, values=values, crs=self.crs,
                           nodata_value=self.nodata)

    def _align_origin(self, x, y, method='nearest'):
        """ Shift the grid so that the transform anchor is an integer multiple
        of x, y from the coordinate system origin at (0, 0).

        For example, if a grid has transform [3.2, 3.2, 1, 1, 0, 0] (raster
        pixels of dimension 1x1, with a lower left outer corner at 3.2, 3.2),
        calling `align_origin(1, 1)` will resample the grid to have transform
        [3, 3, 1, 1, 0, 0], while `align_origin(5, 5)` with resample the grid
        to [5, 5, 1, 1, 0, 0]. In both cases, the grid reference has been
        shifted.

        Arguments
        ---------
        x, y : float
        method : str, optional

        Returns
        -------
        RegularGrid
        """
        t = [x * round(self._transform[0] / x),
             y * round(self._transform[1] / y),
             self._transform[2],
             self._transform[3],
             self._transform[4],
             self._transform[5]]
        return self._resample_transform(t, method=method)

    def resample(self, dx, dy, method='nearest'):
        """ Resample array to have spacing *dx*, *dy*. The grid origin remains
        in the same position.

        Parameters
        ----------
        dx : float
            cell dimension 1
        dy : float
            cell dimension 2
        method : str, optional
            interpolation method, currently 'nearest' and 'linear' supported
        """
        if dx <= 0 or dy <= 0:
            raise ValueError("resolution must be positive "
                             "(got {0}, {1})".format(dx, dy))

        xmin, xmax, ymin, ymax = self.get_extent()
        ny = int((ymax - ymin) // dy) + 1
        nx = int((xmax - xmin) // dx) + 1

        t = self._transform
        tnew = (xmin-0.5*dx-0.5*t[4], ymin-0.5*dy-0.5*t[5], dx, dy, t[4], t[5])
        cg = CoordinateGenerator(tnew, (ny, nx), self.crs, self.crs)
        X, Y = cg[:,:]

        if method == 'nearest':
            values = self.sample_nearest(X, Y)
        elif method == 'linear':
            values = self.sample_bilinear(X, Y)
        else:
            raise NotImplementedError('method "{0}" unavailable'.format(method))

        if values.ndim == 3:
            values = values.transpose(1, 2, 0)
        return RegularGrid(tnew, values=values, crs=self.crs,
                           nodata_value=self.nodata)

    def get_positions(self, x, y):
        """ Return the float row and column indices for the point nearest
        geographical coordinates.

        Parameters
        ----------
        x, y : float or vector
            vertices of points to compute indices for

        Raises
        ------
        ValueError
            if inputs have unequal size
        """
        if hasattr(x, "__iter__"):
            x = np.array(x, dtype=np.float64)
            y = np.array(y, dtype=np.float64)
            if len(x) != len(y):
                raise ValueError("inputs must have equal size")
            return crfuncs.get_positions_vec(self.transform, x, y)
        else:
            x = np.array([x], dtype=np.float64)
            y = np.array([y], dtype=np.float64)
            I, J = crfuncs.get_positions_vec(self.transform, x, y)
            return I[0], J[0]

    def _get_indices(self, x, y):
        """ Return the integer row and column indices for the point nearest
        geographical coordinates. Unlike `get_positions()`, this method raises
        and exception when points are out of range.

        Parameters
        ----------
        x, y : float or vector
            vertices of points to compute indices for
        """
        i, j = self.get_positions(x, y)
        i = np.round(i).astype(np.int32)
        j = np.round(j).astype(np.int32)
        return i,j

    def get_indices(self, x, y):
        """ Return the integer row and column indices for the point nearest
        geographical coordinates. Unlike `get_positions()`, this method raises
        and exception when points are out of range.

        Parameters
        ----------
        x, y : float or vector
            vertices of points to compute indices for

        Raises
        ------
        IndexError
            points outside of Grid bbox
        """
        ny, nx = self.size
        i, j = self._get_indices(x, y)
        if hasattr(i, "__iter__"):
            if i.min() < 0 or i.max() > ny-1 or j.min() < 0 or j.max() > nx-1:
                raise IndexError("coordinates outside grid region "
                                 "({0})".format(self.bbox))
        else:
            if i < 0 or i > ny-1 or j < 0 or j > nx-1:
                raise IndexError("coordinate outside grid region "
                                 "({0})".format(self.bbox))
        return i,j

    def sample_nearest(self, x, y):
        """ Return the value nearest to coordinates using a nearest grid center
        sampling scheme.

        Parameters
        ----------
        x, y : float or vector
            vertices of points to compute indices for

        Returns
        -------
        float or vector
            the size of the output has one more dimension than the input. if
            the input are scalar, the output is a vector length p representing
            band values. if the input is a vector length n, the ouput is p x n.
            if the input is an m x n array, the output is p x m x n.

        Raises
        ------
        IndexError
            points outside of Grid bbox
        """
        m, n = self.size

        if not hasattr(x, "__iter__"):
            i, j = self._get_indices(x, y)
            if (0 <= i < m) and (0 <= j < n):
                return np.atleast_1d(self[i,j])
            else:
                return self.nodata_value

        if not isinstance(x, np.ndarray):
            x = np.array(x)
            y = np.array(y)

        I, J = self._get_indices(x.ravel(), y.ravel())
        out_of_bounds_mask = (I < 0) | (I >= m) | (J < 0) | (J >= n)
        I = I[~out_of_bounds_mask]
        J = J[~out_of_bounds_mask]

        Imn = int(np.floor(I.min())-1)
        Imx = int(np.ceil(I.max()+1))
        Jmn = int(np.floor(J.min())-1)
        Jmx = int(np.ceil(J.max()+1))

        Imn = max(Imn, 0)
        Jmn = max(Jmn, 0)
        Imx = min(Imx, self.size[0])
        Jmx = min(Jmx, self.size[1])

        I -= Imn
        J -= Jmn

        data = np.empty((out_of_bounds_mask.shape[0], self.nbands),
                        dtype=self.bands[0].dtype)

        data[out_of_bounds_mask, :] = self.nodata_value
        data[~out_of_bounds_mask, :] = self[Imn:Imx,Jmn:Jmx,:][I,J].reshape(sum(~out_of_bounds_mask), self.nbands)

        if x.ndim == 1:
            return data.reshape(self.nbands, x.shape[0])
        else:
            return data.reshape(self.nbands, x.shape[0], x.shape[1])

    def sample_bilinear(self, x, y):
        """ Return the value nearest to coordinates using a bi-linear sampling
        scheme.

        Parameters
        ----------
        x, y : float or vector
            vertices of points to compute indices for

        Returns
        -------
        float or vector
            the size of the output has one more dimension than the input. if
            the input are scalar, the output is a vector length p representing
            band values. if the input is a vector length n, the ouput is p x n.
            if the input is an m x n array, the output is p x m x n.

        Raises
        ------
        IndexError
            points outside of Grid bbox
        """
        if not hasattr(x, "__iter__"):
            dim = 0
            x = np.array([x])
            y = np.array([y])
        elif not isinstance(x, np.ndarray):
            x = np.array(x)
            y = np.array(y)
            dim = x.ndim
        else:
            dim = x.ndim

        I, J = self.get_positions(x.ravel(), y.ravel())

        # If the grid is large, decompressing everthing is expensive, so
        # compute only the region necessary
        if len(I) != 1:
            Imn = int(np.floor(I.min())-1)
            Imx = int(np.ceil(I.max()+1))
            Jmn = int(np.floor(J.min())-1)
            Jmx = int(np.ceil(J.max()+1))

            Imn = max(Imn, 0)
            Jmn = max(Jmn, 0)
            Imx = min(Imx, self.size[0])
            Jmx = min(Jmx, self.size[1])

            I -= Imn
            J -= Jmn
        else:
            Imn = None
            Imx = None
            Jmn = None
            Jmx = None

        data = []
        for i, band in enumerate(self.bands):
            v = self[Imn:Imx,Jmn:Jmx,i].squeeze()
            if band.dtype in (np.float32, np.float64):
                data.append(crfuncs.sample_bilinear_double(I, J, v.astype(np.float64), self.nodata_value))
            elif band.dtype in (np.int16, np.int32, np.int64):
                data.append(crfuncs.sample_bilinear_int(I, J, v.astype(np.int32), self.nodata_value))
            elif band.dtype in (np.uint8, np.uint16, np.uint32):
                data.append(crfuncs.sample_bilinear_uint(I, J, v.astype(np.uint16), self.nodata_value))
            else:
                raise NotImplementedError("no sample_bilinear method for dtype:"
                                          " {0}".format(band.dtype))

        if dim == 0:
            return np.array([d[0] for d in data])
        elif dim == 1:
            return np.atleast_2d(data)
        elif dim == 2:
            m, n = x.shape
            return np.concatenate([d.reshape((1, m, n)) for d in data], axis=0)

    def sample(self, *args, **kwargs):
        """ Return the values nearest positions. Positions may be:

        - a karta.Point instance
        - a karta.Multipoint instance, or
        - a pair of x, y coordinate arrays

        Parameters
        ----------
        positions : karta.Point, karta.Multipoint, or two lists of floats
            see above
        crs : karta.crs.CRS, optional
            used when coordinate lists are provided, otherwise the coordinate
            system is taken from the crs attribute of the geometry
        method : string, optional
            may be one of 'nearest', 'bilinear' (default).

        Returns
        -------
        float or vector
            The size of the output has one more dimension than the input. If
            the input are scalar, the output is a p-vector representing band
            values. If the input is a n-vector, the ouput is p x n. If the
            input is an m x n array, the output is p x m x n.

        Raises
        ------
        IndexError
            points outside of Grid bbox
        """
        crs = kwargs.get("crs", None)
        method = kwargs.get("method", "bilinear")

        argerror = TypeError("'sample' method takes a Point, a Multipoint, or x and y coordinate arrays")
        if hasattr(args[0], "_geotype"):
            crs = args[0].crs
            if args[0]._geotype == "Point":
                x, y = args[0].get_vertex(crs=self.crs)[:2]
            elif args[0]._geotype == "Multipoint":
                x, y = args[0].get_coordinate_lists(crs=self.crs)
            else:
                raise argerror
        else:
            try:
                x = args[0]
                y = args[1]

                if crs is None:
                    crs = self.crs
                else:
                    x, y = crs.transform(self.crs, x, y)
            except IndexError:
                raise argerror
            if len(x) != len(y):
                raise argerror

        if method == "nearest":
            v = self.sample_nearest(x, y)
        elif method == "bilinear":
            v = self.sample_bilinear(x, y)
        else:
            raise ValueError("method '{0}' not available".format(method))
        return v

    def profile(self, line, resolution=None, **kw):
        """ Sample along a *Line* at a specified interval.

        Parameters
        ----------
        line : karta.Line
            defines the sampling path
        resolution : float, optional
            sample spacing, taken to be the minimum grid resolution by default

        Additional keyword arguments passed to `RegularGrid.sample` (e.g. to
        specify sampling method)

        Returns
        -------
        Multipoint
            sample points
        ndarray
            grid value at sample points
        """
        if resolution is None:
            resolution = min(self.transform[2:4])
        points = line.to_points(resolution)
        z = self.sample(*points.coordinates, **kw)
        if len(self.bands) != 1:
            for i in range(self.nbands):
                points.data.setfield("band_{0}".format(i), z[i])
        else:
            points.data.setfield("band_0", z[0])
        return points, z

    def to_geotiff(self, fnm, compress="PACKBITS", tiled=False, **kw):
        """ Write data to a GeoTiff file using GDAL.

        Parameters
        ----------
        fnm : str
            output file name
        compress: str or None, optional
            "PACKBITS" (default), "DEFLATE", "LZW", "LZMA", or None
        """
        return _gdal.write(fnm, self, compress=compress, **kw)

    def to_gtiff(self, *args, **kwargs):
        """ Alias for to_geotiff """
        warnings.warn("method `to_gtiff` has been renamed `to_geotiff`",
                FutureWarning)
        return self.to_geotiff(*args, **kwargs)

    def to_aai(self, f, reference='corner', nodata_value=-9999):
        """ Save internal data as an ASCII grid. Based on the ESRI standard,
        only isometric grids (i.e. `hdr['dx'] == hdr['dy']` can be saved.

        Parameters
        ----------
        f : str
            a file-like object or a filename
        reference : str
            specify a header reference ('center' | 'corner')
        nodata_value : number
            specify how NaNs should be represented

        Raises
        ------
        GridError
            input grid is not isometric and can therefore not be saved
        """
        if reference not in ('center', 'corner'):
            raise errors.GridError("reference in AAIGrid.tofile() must be 'center' or "
                           "'corner'")

        dx, dy = self.resolution
        if dx != dy:
            raise errors.GridError("ESRI ASCII grids require isometric grid cells")

        sx, sy = self.skew
        if sx != sy != 0.0:
            raise errors.GridError("skewed grids cannot be written as ESRI ASCII")

        ny, nx = self.bands[0].size

        if not hasattr(f, 'read'):
            f = open(f, "w")

        try:
            data_a = self[:,:].copy()
            data_a[np.isnan(data_a)] = nodata_value

            f.write("NCOLS {0}\n".format(nx))
            f.write("NROWS {0}\n".format(ny))
            if reference == 'center':
                xll, yll = self.center_llref()
                f.write("XLLCENTER {0}\n".format(xll))
                f.write("YLLCENTER {0}\n".format(yll))
            elif reference == 'corner':
                xll, yll = self.corner_llref()
                f.write("XLLCORNER {0}\n".format(xll))
                f.write("YLLCORNER {0}\n".format(yll))
            f.write("CELLSIZE {0}\n".format(dx))
            f.write("NODATA_VALUE {0}\n".format(nodata_value))
            f.writelines([str(row).replace(',','')[1:-1] +
                            '\n' for row in data_a.tolist()])
        finally:
            f.close()
        return

def merge(grids, weights=None):
    """ Construct a grid mosiac by averaging multiple grids. Currently limited
    to grids whose sampling is an integer translation from each other.

    Parameters
    ----------
    grids : iterable of Grid objects
        grids to combine
    weights : iterable of floats, optional
        weighting factors for computing grid averages

    Raises
    ------
    GridError, ValueError
        input grids have inconsistent transform properties
    """

    # Check grid class
    if not all(isinstance(grid, RegularGrid) for grid in grids):
        raise TypeError("all grids must be type RegularGrid")

    T = grids[0].transform
    # Check grid stretch and skew
    for i, grid in enumerate(grids[1:]):
        if grid.transform[2:6] != T[2:6]:
            raise errors.GridError("grid %d transform stretch and skew "
                                   "does not match grid 1" % (i+2,))

    # Check grid offset
    excmsg = "grid %d not an integer translation from grid 1"
    for i, grid in enumerate(grids[1:]):
        if ((grid.transform[0]-T[0]) / float(T[2])) % 1 > 1e-15:
            raise ValueError(excmsg % (i+2,))
        if ((grid.transform[1]-T[1]) / float(T[3])) % 1 > 1e-15:
            raise ValueError(excmsg % (i+2,))

    # Check number of grid bands
    if len(set([len(grid.bands) for grid in grids])) != 1:
        raise ValueError("all grids to mosaic must have the same number of bands")

    # Compute weighting coefficients by normalizing *weights*
    if weights is None:
        weights = np.ones(len(grids))
    else:
        weights = np.asarray(weights, dtype=np.float32)

    normalizedweights = weights * len(weights) / weights.sum()

    # Compute final grid extent
    xmin, xmax, ymin, ymax = grids[0].get_extent(reference='edge')
    for grid in grids[1:]:
        _xmin, _xmax, _ymin, _ymax = grid.get_extent(reference='edge')
        xmin = min(xmin, _xmin)
        xmax = max(xmax, _xmax)
        ymin = min(ymin, _ymin)
        ymax = max(ymax, _ymax)

    nx = int(round((xmax-xmin) / T[2]))
    ny = int(round((ymax-ymin) / T[3]))

    # Allocate data array and copy each grid's data
    typ = grids[0].bands[0].dtype
    outbands = []

    for iband in range(len(grids[0].bands)):
        values = np.zeros([ny, nx], dtype=typ)
        counts = np.zeros([ny, nx], dtype=np.float32)

        for grid, weight in zip(grids, normalizedweights):
            _xmin, _xmax, _ymin, _ymax = grid.get_extent(reference='edge')
            offx = int((_xmin-xmin) / T[2])
            offy = int((_ymin-ymin) / T[3])
            _ny, _nx = grid.size

            mask = grid.data_mask
            counts[offy:offy+_ny,offx:offx+_nx][mask] += weight
            _bnd = grid.bands[iband]
            values[offy:offy+_ny,offx:offx+_nx][mask] += typ(_bnd.getblock(0, 0, *_bnd.size)[mask]) * weight
            del mask

        validcountmask = (counts!=0.0)
        values[validcountmask] = values[validcountmask] / counts[validcountmask]
        values[~validcountmask] = grids[0].nodata

        band = CompressedBand((ny, nx), typ, initval=grids[0].nodata)
        band.setblock(0, 0, values)
        outbands.append(band)

    Tmerge = [xmin, ymin] + list(T[2:])
    return RegularGrid(Tmerge, bands=outbands, crs=grids[0].crs,
                       nodata_value=grids[0].nodata)

def get_nodata(T):
    """ Return a default value for NODATA given a type

    For unsigned integer types, returns largest representable value.
    For signed integer types, returns smallest negative representable value.
    For floating point types (incl. complex), returns NaN.

    Parameters
    ----------
    T : type
        numeric type to choose NODATA for

    Raises
    ------
    ValueError
        input is not an integer, real, or complex numeric type
    """
    if T in (np.uint8, np.uint16, np.uint32, np.uint64):
        return np.iinfo(T).max
    elif issubclass(T, numbers.Integral):
        return np.iinfo(T).min
    elif issubclass(T, (numbers.Real, numbers.Complex)):
        return np.nan
    else:
        raise ValueError("No default NODATA value for type {0}".format(T))

def gridpoints(x, y, z, transform, crs):
    """ Return a grid computed by averaging point data over cells.

    Parameters
    ----------
    x : iterable
        point data x coordinates
    y : iterable
        point data y coordinates
    z : iterable
        point data values
    transform : 6-tuple of floats
        geotransform: ``[xllcorner, yllcorner, xres, yres, xskew, yskew]``
    crs : karta.crs.CRS subclass
        coordinate reference system object
    """
    ny = int((np.max(y) - transform[1]) // transform[3]) + 1
    nx = int((np.max(x) - transform[0]) // transform[2]) + 1
    grid = RegularGrid(transform, values=np.empty([ny, nx]), crs=crs,
                       nodata_value=np.nan)

    array = np.zeros([ny, nx])
    (I, J) = grid.get_indices(x, y)

    try:
        err = crfuncs.fillarray_double(array,
                                       I.astype(np.int32),
                                       J.astype(np.int32), z, grid.nodata)
        if err != 0:
            raise RuntimeError("failure in fillarray_double")
    except ValueError:
        # Fast version works when *z* is of type double (np.float64).
        # Python fallback for other types
        counts = np.zeros([ny, nx], dtype=np.int16)
        for (i,j,z_) in zip(I, J, z):
            array[i,j] += z_
            counts[i,j] += 1

        m = counts!=0
        array[m] = array[m] / counts[m]
        array[~m] = np.nan

    grid[:,:,0] = array
    return grid

def mask_poly(xpoly, ypoly, nx, ny, transform):
    """ Create a grid mask based on a clockwise-oriented polygon.

    Parameters
    ----------
    xpoly, ypoly : lists of floats
        sequences of points representing polygon
    nx : int
    ny : int
        size of grid
    transform : list[float]
        affine transformation describing grid layout and origin
        ``T == [x0, y0, dx, dy, sx, sy]``

    Returns
    -------
    ndarray
    """
    mask = np.zeros((ny, nx), dtype=np.int8)

    # find left southernmost index (will start and end on this point)
    x0, y0 = xpoly[0], ypoly[0]
    i_bot = 0
    for i in range(1, len(ypoly)):
        if ypoly[i] < y0 or (ypoly[i] == y0 and xpoly[i] < x0):
            x0 = xpoly[i]
            y0 = ypoly[i]
            i_bot = i

    x0 = xpoly[i_bot]
    y0 = ypoly[i_bot]

    # compute the grid indices of the starting point
    ta, tb, tc, td, te, tf = transform

    # normalize grid transform so increasing i, j corresponds to north, east
    if tc < 0:
        ta = ta + nx*tc
        tc = -tc

    if td < 0:
        tb = tb + ny*td
        td = -td

    i0 = int(round((y0-tb - tf/tc*(x0-ta)) / (td - tf*te/tc)))
    j0 = int(round((x0-ta - te/td*(y0-tb)) / (tc - te*tf/td)))

    for el in range(1, len(xpoly)+1):
        idx = (el + i_bot) % len(xpoly)
        x1 = xpoly[idx]
        y1 = ypoly[idx]

        # (Unbounded) grid indices of the segment end points
        i1 = int(round((y1-tb - tf/tc*(x1-ta)) / (td - tf*te/tc)))
        j1 = int(round((x1-ta - te/td*(y1-tb)) / (tc - te*tf/td)))

        # If segment is horizontal or off-grid, ignore
        if ((0 <= i0 < ny) and (0 <= i1 < ny)) or (y1 != y0):

            if y1 > y0:     # upward - mark grid cells to the right

                for i in range(i0, i1):
                    if (0 <= i < ny):
                        j = int(round((i-i0) * (x1-x0)/(y1-y0) + j0))
                        if j < nx:
                            mask[i, max(0, j):] += 1

            else:           # downward - unmark grid cells to the right

                for i in range(i1, i0):
                    if (0 <= i < ny):
                        j = int(round((i-i1) * (x1-x0)/(y1-y0) + j1))
                        if j < nx:
                            mask[i, max(0, j):] -= 1

        x0 = x1
        y0 = y1
        i0 = i1
        j0 = j1

    return mask.astype(np.bool)




"""
Karta is a collection of Python modules for handling vector and raster
geospatial data.
"""

__all__ = ["vector", "raster", "crs", "tile", "errors"]

from .version import __version__

from . import vector
from . import raster
from . import crs
from . import tile
from . import errors

from .vector import (Point, Line, Polygon, Multipoint, Multiline, Multipolygon,
                     from_shape, read_geojson,
                     read_gpx_waypts, read_gpx_tracks,
                     read_shapefile, write_shapefile,
                     geometry)

from .raster import (RegularGrid, SimpleBand, CompressedBand,
                     read_aai, read_geotiff, read_gtiff, from_geotiffs,
                     grid, misc)

# from .tile import tile_bbox, tile_nw_corner, tile_tuple




from libc.math cimport M_PI, M_PI_2
from cpython cimport bool
from coordstring cimport CoordString
from vectorgeo cimport Vector2, azimuth, azimuth_sph, dist2, dist_sph

cdef bool isleft(Vector2 pt0, Vector2 pt1, Vector2 pt2):
    """ Indicate whether pt0 is to the left of a segment between pt1 and pt2.
    """
    return (pt1.x-pt0.x)*(pt2.y-pt0.y) - (pt1.y-pt0.y)*(pt2.x-pt0.x) > 0.0

cdef bool isleft_sph(Vector2 pt0, Vector2 pt1, Vector2 pt2):
    cdef double az, az_pt, daz
    az = azimuth_sph(pt1, pt2)
    az_pt = azimuth_sph(pt1, pt0)
    daz = ((az-az_pt) + M_PI) % (2*M_PI) - M_PI
    return daz > 0

def convexhull(CoordString cs):
    """ Return the convex hull for coordinates on a plane. """
    # Find the leftmost (upper) point
    cdef int n = len(cs)
    if cs.ring:
        n += 1
    cdef int ileftmost = 0
    cdef double x, xleftmost = cs.getX(0)
    cdef int i = 1

    while i != n:
        x = cs.getX(i)
        if x < xleftmost:
            xleftmost = x
            ileftmost = i
        elif x == xleftmost and (cs.getY(i) > cs.getY(ileftmost)):
            ileftmost = i
        i += 1

    # Sort CCW relative to pt0
    cdef list azimuth_indices = []
    cdef double az, y
    cdef Vector2 leftmost = Vector2(xleftmost, cs.getY(ileftmost))
    i = 0
    while i != n:
        if i == ileftmost:
            i += 1
            continue
        az = M_PI_2 - azimuth(leftmost, Vector2(cs.getX(i), cs.getY(i)))
        azimuth_indices.append((az, i))
        i += 1
    azimuth_indices.sort()

    # Drop all but farthest of any duplicates
    cdef int last_idx = 0
    cdef Vector2 ipt, lastpt = leftmost
    cdef list indices = [azimuth_indices[0][1]]
    i = 1
    while i != len(azimuth_indices):
        if azimuth_indices[i][0] == azimuth_indices[last_idx][0]:
            ipt = Vector2(cs.getX(azimuth_indices[i][1]), cs.getY(azimuth_indices[i][1]))
            if dist2(leftmost, ipt) > dist2(leftmost, lastpt):
                # replace last index with that of the current vertex
                indices[-1] = azimuth_indices[i][1]
                last_idx = i
                lastpt = Vector2(cs.getX(last_idx), cs.getY(last_idx))
        else:
            indices.append(azimuth_indices[i][1])
        i += 1

    # Initialize and scan for convex hull
    if len(indices) == 1:
        return [ileftmost, indices[0]]

    cdef list hull_indices = [ileftmost, indices[0], indices[1]]
    if len(indices) == 2:
        return hull_indices

    for i in indices[2:]:
        while not isleft(Vector2(cs.getX(hull_indices[-2]), cs.getY(hull_indices[-2])),
                         Vector2(cs.getX(hull_indices[-1]), cs.getY(hull_indices[-1])),
                         Vector2(cs.getX(i), cs.getY(i))):
            hull_indices.pop()
        hull_indices.append(i)
    return hull_indices

def convexhull_sph(CoordString cs):
    """ Return the convex hull for coordinates on a sphere. """
    # Find the leftmost (upper) point
    cdef int n = len(cs)
    if cs.ring:
        n += 1
    cdef int ileftmost = 0
    cdef double x, xleftmost = cs.getX(0)
    cdef int i = 1

    while i != n:
        x = cs.getX(i)
        if x < xleftmost:
            xleftmost = x
            ileftmost = i
        elif x == xleftmost and (cs.getY(i) > cs.getY(ileftmost)):
            ileftmost = i
        i += 1

    # Sort CCW relative to pt0
    cdef list azimuth_indices = []
    cdef double az, y
    cdef Vector2 leftmost = Vector2(xleftmost, cs.getY(ileftmost))
    i = 0
    while i != n:
        if i == ileftmost:
            i += 1
            continue
        az = M_PI_2 - azimuth_sph(leftmost, Vector2(cs.getX(i), cs.getY(i)))
        azimuth_indices.append((az, i))
        i += 1
    azimuth_indices.sort()

    # Drop all but farthest of any duplicates
    cdef int last_idx = 0
    cdef Vector2 ipt, lastpt = leftmost
    cdef list indices = [azimuth_indices[0][1]]
    i = 1
    while i != len(azimuth_indices):
        if azimuth_indices[i][0] == azimuth_indices[last_idx][0]:
            ipt = Vector2(cs.getX(azimuth_indices[i][1]), cs.getY(azimuth_indices[i][1]))
            if dist_sph(leftmost, ipt) > dist_sph(leftmost, lastpt):
                # replace last index with that of the current vertex
                indices[-1] = azimuth_indices[i][1]
                last_idx = i
                lastpt = Vector2(cs.getX(last_idx), cs.getY(last_idx))
        else:
            indices.append(azimuth_indices[i][1])
        i += 1

    # Initialize and scan for convex hull
    if len(indices) == 1:
        return [ileftmost, indices[0]]

    cdef list hull_indices = [ileftmost, indices[0], indices[1]]
    if len(indices) == 2:
        return hull_indices

    for i in indices[2:]:
        while not isleft_sph(Vector2(cs.getX(hull_indices[-2]), cs.getY(hull_indices[-2])),
                             Vector2(cs.getX(hull_indices[-1]), cs.getY(hull_indices[-1])),
                             Vector2(cs.getX(i), cs.getY(i))):
            hull_indices.pop()
        hull_indices.append(i)
    return hull_indices

def convexhull_simple(CoordString cs):
    """ Return the convex hull for a simple polyline on a plane. """
    raise NotImplementedError()



import math
from .vector.geometry import Point
from .crs import LonLatWGS84

class Tile(object):

    def __init__(self, z, x, y):
        self.z = z
        self.x = x
        self.y = y

    def __eq__(self, other):
        if not isinstance(other, Tile):
            return False
        return (self.z == other.z) and (self.x == other.x) and (self.y == other.y)
    
    def __neq__(self, other):
        return not (self == other)

    def nw_corner(self):
        """ Return the northwest corner point from a Tile.

        Returns
        -------
        Point
        """
        z = self.z
        x = self.x
        y = self.y
        n = 2**z
        lon = float(x)/n * 360.0 - 180.0
        lat = math.atan(math.sinh(math.pi * (1-2*y/n))) * 180.0 / math.pi
        return Point((lon, lat), crs=LonLatWGS84)

    @property
    def bbox(self):
        """ Return a tuple representing the bounding box of a Tile

        Returns
        -------
        tuple
            (xmin, ymin, xmax, ymax)
        """
        z = self.z
        x = self.x
        y = self.y
        pts = [self.nw_corner(),
               Tile(z, x+1, y).nw_corner(),
               Tile(z, x, y+1).nw_corner()]
        return pts[0].x, pts[2].y, pts[1].x, pts[0].y

def tile_from_point(point, zoom):
    """ Return the (z, x, y) locator for an OpenStreetMap tile containing a
    point.

    Parameters
    ----------
    point : Point
        geographic point to be contained within tile
    zoom : int
        non-negative zoom level (typically 0-18)

    Returns
    -------
    Tile
    """
    z = int(zoom)
    dlon = 256
    dlat = 256

    lon0, lat0 = point.crs.project(*point.vertex[:2], inverse=True)
    c = 128/math.pi * 2**z
    x0 = c * (lon0*math.pi/180+math.pi)
    y0 = c * (math.pi-math.log(math.tan(math.pi/4+lat0*math.pi/360)))

    x = int(x0 // dlon)
    y = int(y0 // dlat)
    return Tile(z, x, y)



import numpy as np
cimport numpy as np
from cpython cimport bool
from cpython.array cimport array, clone

cdef double mind(double a, double b) nogil:
    return a if a<= b else b

cdef double maxd(double a, double b) nogil:
    return a if a>= b else b

cdef int floord(double a) nogil:
    if (a % 1) == 0:
        return <int> a
    else:
        return <int> (a // 1)

cdef int ceild(double a) nogil:
    if (a % 1) == 0:
        return <int> a
    else:
        return floord(a) + 1

cdef array template_dbl = array("d", [])

cdef class CoordString:
    """ A CoordString is a datastructure for coordinate data. Initialize a
    CoordString with 2 or 3 iterable objects, corresponding to x, y, and
    optionally z coordinates.

    Parameters
    ----------
    coords : iterable
        list of coordinates, with items of length 2 or 3
    """
    def __cinit__(self, object coords, bool ring=False):
        cdef int length = -1
        try:
            length = len(coords)
            self.length = length
            if length == 0:
                self.rank = -1
            else:
                self.rank = len(coords[0])
        except TypeError:
            raise ValueError("coords argument must be an iterable of iterables")

        if self.rank not in (-1, 2, 3):
            raise ValueError("CoordString rank must be 2 or 3")

        self.ring = ring

        cdef array coords_array = clone(template_dbl, length*self.rank, False)
        self.coords = coords_array

        cdef int i = 0, j
        cdef object xy
        for xy in coords:
            for j in range(self.rank):
                self.coords[i+j] = xy[j]
            i = i + self.rank
        return

    def __len__(self):
        if self.rank == -1:
            return 0
        else:
            return self.length

    def __iter__(self):
        cdef int i = 0
        while i < self.length:
            yield self[i]
            i += 1

    def __getitem__(self, int index):
        cdef double x, y, z
        cdef int pos = index*self.rank
        if self.rank == -1:
            raise IndexError("CoordString empty")
        x = self.coords[pos]
        y = self.coords[pos+1]
        if self.rank == 2:
            return x, y
        else:
            z = self.coords[pos+2]
            return x, y, z

    def __setitem__(self, int key, double[:] value):
        cdef int idx = key*self.rank
        if self.rank == -1:
            raise IndexError("CoordString empty")
        self.coords[key*self.rank:(key+1)*self.rank] = value

    def __hash__(self):
        return hash(self.coords)

    def __richcmp__(self, other, int op):
        if op == 2:
            return self._ceq(other)
        elif op == 3:
            return not self._ceq(other)
        else:
            raise NotImplementedError("CoordString comparison")

    def _ceq(self, CoordString other):
        """ Tests equality """
        cdef int i = 0
        if self.length != other.length:
            return False
        for i in range(self.length):
            if self.coords[i] != other.coords[i]:
                return False
        return True

    cdef double getX(self, int index):
        if self.ring and index == self.length:
            index = 0
        return self.coords[index*self.rank]

    cdef double getY(self, int index):
        if self.ring and index == self.length:
            index = 0
        return self.coords[index*self.rank+1]

    cdef double getZ(self, int index):
        if self.rank != 3:
            return np.nan
        if self.ring and index == self.length:
            index = 0
        return self.coords[index*self.rank+2]

    cpdef np.ndarray slice(self, int start, int stop=0, int step=1):
        """ Slice coordinate string, returning an <n x rank> numpy array. """
        while start < 0:
            start += self.length
        while stop <= 0:
            stop += self.length

        cdef int outlength
        if step != 0:
            outlength = ceild(<double> (abs(stop) - abs(start)) / abs(step))
        else:
            raise ValueError("step cannot equal zero")

        if self.rank == -1:
            return ValueError("CoordString empty")

        cdef np.ndarray result = np.empty([outlength, self.rank], dtype=np.float64)
        cdef int i = 0, pos = start
        cdef int rank = self.rank
        while pos < stop:
            result[i,:] = self.coords[pos*rank:(pos+1)*rank]
            i += 1
            pos += step
        return result

    @property
    def bbox(self):
        cdef double xmin, ymin, xmax, ymax
        cdef int i = 0
        cdef int offset = 0

        if self.length == 0:
            return (np.nan, np.nan, np.nan, np.nan)

        xmin = self.coords[0]
        xmax = self.coords[0]
        ymin = self.coords[1]
        ymax = self.coords[1]
        while i != self.length-1:
            offset += self.rank
            xmin = mind(xmin, self.coords[offset])
            xmax = maxd(xmax, self.coords[offset])
            ymin = mind(ymin, self.coords[offset+1])
            ymax = maxd(ymax, self.coords[offset+1])
            i += 1
        return (xmin, ymin, xmax, ymax)

    def vectors(self, bool drop_z=False):
        cdef int i = 0, pos = 0
        cdef np.ndarray x, y
        x = np.empty(self.length, dtype=np.float64)
        y = np.empty(self.length, dtype=np.float64)
        while i < self.length:
            x[i] = self.coords[pos]
            y[i] = self.coords[pos+1]
            i += 1
            pos += self.rank
        if (self.rank == 2) or drop_z:
            return x, y

        cdef np.ndarray z
        z = np.empty(self.length, dtype=np.float64)
        pos = 0
        i = 0
        while i < self.length:
            z[i] = self.coords[pos+2]
            i += 1
            pos += self.rank
        return x, y, z

    def asarray(self):
        if self.rank == -1:
            return np.array([[]], dtype=np.float64)
        return np.array(self.coords).reshape([-1, self.rank])



from libc.math cimport fabs
from coordstring cimport CoordString
from vectorgeo cimport Vector2

cdef double isleft(Vector2 pt, Vector2 pt0, Vector2 pt1):
    """ tests whether *pt* is left of segment (pt0, pt1).

    returns > 0 if left, < 0 if right, and = 0 if pt is on the segment
    """
    return (pt1.x - pt0.x) * (pt.y - pt0.y) - (pt.x - pt0.x) * (pt1.y - pt0.y)

def contains(double x, double y, CoordString poly):
    """ Uses a winding number scheme to compute whether *poly* contains a point
    (x, y) """
    cdef int cnt = 0
    cdef int i
    cdef int n = len(poly)
    cdef Vector2 pt, pt0, pt1

    if not poly.ring:
        raise TypeError("no membership test for non-ring coordinate strings")

    pt = Vector2(x, y)
    pt0 = Vector2(poly.getX(0), poly.getY(0))
    for i in range(n):
        pt1 = Vector2(poly.getX(i+1), poly.getY(i+1))
        if (pt0.y <= pt.y < pt1.y):
            # crosses upward
            if isleft(pt, pt0, pt1) > 0:
                cnt += 1
        elif (pt0.y > pt.y >= pt1.y):
            # crosses downward
            if isleft(pt, pt0, pt1) < 0:
                cnt -= 1
        pt0 = pt1

    return cnt != 0

def contains_proj(double x, double y, CoordString poly, crs):
    """ contains implementation for geographical coordinates.
    calls crs.inverse n times, making this relatively inefficient.
    """
    cdef double sum_az = 0.0, az0, az1
    cdef double x0, y0, x1, y1
    cdef int i

    if not poly.ring:
        raise TypeError("no membership test for non-ring coordinate strings")

    x0 = poly.getX(0)
    y0 = poly.getY(0)
    az0, _, _ = crs.inverse(x, y, x0, y0)
    az0 = (az0 + 360 ) % 360
    for i in range(len(poly)):
        x1 = poly.getX(i+1)
        y1 = poly.getY(i+1)

        az1, _, _ = crs.inverse(x, y, x1, y1)
        az1 = (az1 + 360 ) % 360

        # correct for crossing prime meridian
        if az1 - az0 < -180:
            # ccw
            sum_az += (az1-az0) + 360
        elif az0 - az1 > 180:
            # cc
            sum_az += (az1-az0) - 360
        else:
            sum_az += (az1-az0)

        az0 = az1
        x0 = x1
        y0 = y1

    if fabs(sum_az) > 1e-4:
        return False
    else:
        return True




"""
Band implementations for storing data in Karta Grid instances

Overview
--------

`BandIndexer` interface for accessing data from one or more bands

`SimpleBand` use numpy arrays for data storage

`CompressedBand` uses blosc compression to reduce in-memory footprint

Implementation
--------------

Bands are expected to implement the following methods:
    - `__init__(self, size, dtype, initval=None)`
    - `getblock(self, yoff, xoff, ny, nx)`
    - `setblock(self, yoff, xoff, array)`

Attributes:
    - `dtype`
    - `size`

The following methods are deprecated:
    - `__getitem__(self, key)`, accepting as *key* any of
        - an int
        - a slice
        - a 2-tuple of ints
        - a 2-tuple of slices
    - `__setitem__(self, key, value)`, accepting as *key* the same
      possibilities as __getitem__
"""

import blosc
import numpy as np
from numbers import Real, Integral
from math import ceil

class BandIndexer(object):

    def __init__(self, bands):
        self.bands = bands

    def __getitem__(self, key):
        if isinstance(key, np.ndarray):
            return self._get_from_array_mask(key)

        if isinstance(key, slice):
            key = (key, slice(None, None, None), slice(None, None, None))

        if not isinstance(key, tuple):
            raise TypeError("key should be an array or a tuple")

        collapse_rows = collapse_cols = collapse_bands = False
        ny, nx = self.bands[0].size

        if isinstance(key[0], Integral):
            collapse_rows = True
            r = key[0] % ny
            ystart, yend, ystep = (r, r+1, 1)
        elif isinstance(key[0], slice):
            ystart, yend, ystep = key[0].indices(ny)
        else:
            raise TypeError("first key item should be an integer or a slice")

        if isinstance(key[1], Integral):
            collapse_cols = True
            r = key[1] % nx
            xstart, xend, xstep = (r, r+1, 1)
        elif isinstance(key[1], slice):
            xstart, xend, xstep = key[1].indices(nx)
        else:
            raise TypeError("second key item should be an integer or a slice")

        if len(key) == 2:
            bands = list(range(len(self.bands)))
        elif len(key) == 3 and isinstance(key[2], Integral):
            collapse_bands = True
            bands = [key[2] % len(self.bands)]
        elif len(key) == 3 and isinstance(key[2], slice):
            bands = list(range(*key[2].indices(len(self.bands))))
        else:
            raise TypeError("third key item should be an integer or a slice")

        if ystep < 0:
            ystart, yend = yend+1, ystart+1

        if xstep < 0:
            xstart, xend = xend+1, xstart+1

        shape = [1 + (yend-ystart-1) // abs(ystep),
                 1 + (xend-xstart-1) // abs(xstep),
                 len(bands)]

        out = np.empty(shape, dtype = self.bands[0].dtype)

        for i, iband in enumerate(bands):
            band = self.bands[iband]
            band_values = band.getblock(ystart, xstart, yend-ystart, xend-xstart)
            out[:,:,i] = band_values[::ystep,::xstep]

        if collapse_bands:
            out = out[:,:,0]
        if collapse_cols:
            out = out[:,0]
        if collapse_rows:
            out = out[0]

        return out

    def __setitem__(self, key, value):
        if isinstance(key, np.ndarray):
            return self._set_from_array_mask(key, value)

        if isinstance(key, slice):
            key = (key, slice(None, None, None), slice(None, None, None))

        if not isinstance(key, tuple):
            raise TypeError("key should be an array or a tuple")

        ny, nx = self.bands[0].size

        if isinstance(key[0], Integral):
            r = key[0] % ny
            ystart, yend, ystep = (r, r+1, 1)
        elif isinstance(key[0], slice):
            ystart, yend, ystep = key[0].indices(ny)
        else:
            raise TypeError("first key item should be an integer or a slice")

        if isinstance(key[1], Integral):
            r = key[1] % nx
            xstart, xend, xstep = (r, r+1, 1)
        elif isinstance(key[1], slice):
            xstart, xend, xstep = key[1].indices(nx)
        else:
            raise TypeError("second key item should be an integer or a slice")

        if len(key) == 2:
            bands = list(range(len(self.bands)))
        elif len(key) == 3 and isinstance(key[2], Integral):
            collapse_bands = True
            bands = [key[2] % len(self.bands)]
        elif len(key) == 3 and isinstance(key[2], slice):
            bands = list(range(*key[2].indices(len(self.bands))))
        else:
            raise TypeError("third key item should be an integer or a slice")

        if not (xstep == ystep == 1):
            raise NotImplementedError("setting band values with stepped slices")

            #if ystep < 0:
            #    ystart, yend = yend+1, ystart+1

            #if xstep < 0:
            #    xstart, xend = xend+1, xstart+1

        shape = [1 + (yend-ystart-1) // abs(ystep),
                 1 + (xend-xstart-1) // abs(xstep),
                 len(bands)]

        if isinstance(value, np.ndarray) and (value.ndim == 1) and (shape[0] == shape[1] == 1):
            val_array = np.reshape(np.atleast_3d(value), shape)
        else:
            val_array = np.broadcast_to(np.atleast_3d(value), shape)

        for i, iband in enumerate(bands):
            band = self.bands[iband]
            band.setblock(ystart, xstart, val_array[:,:,i])

        return

    def _get_from_array_mask(self, mask):
        # The mask is assumed to be in (row, column[, band]) order
        # TODO: make this memory efficient
        if mask.ndim == 2:
            return self[:,:,:][mask]
        elif mask.ndim == 3:
            return self[:,:,:][mask]
        else:
            raise IndexError("masking array must have two or three dimensions")

    def _set_from_array_mask(self, mask, value):
        # The mask is assumed to be in (row, column[, band]) order
        # TODO: make this memory efficient
        for i, band in enumerate(self.bands):

            if mask.ndim == 3:
                mask_ = mask[:,:,i]
            else:
                mask_ = mask

            tmp = band.getblock(0, 0, *band.size)
            if isinstance(value, Real) or (value.ndim == 1):
                tmp[mask_] = value
            else:
                tmp[mask_] = value[:,i]
            band.setblock(0, 0, tmp)

    def __iter__(self):
        nx = self.bands[0].size[1]
        for i in range(self.bands[0].size[0]):
            if len(self.bands) == 1:
                yield self.bands[0].getblock(i, 0, 1, nx)
            else:
                yield np.vstack([b.getblock(i, 0, 1, nx) for b in self.bands])

    @property
    def shape(self):
        """ Returns the dimensions of raster bands. If there is a single
        (m x n) band, output is a tuple (m, n). If there are N>1 bands, output
        is a tuple (N, m, n).
        """
        if len(self.bands) == 0:
            raise ValueError("no bands")
        else:
            return self.bands[0].size

    @property
    def dtype(self):
        """ Returns bands' dtype """
        return self.bands[0].dtype

class SimpleBand(object):
    """ SimpleBand wraps a numpy.ndarray for storage. """

    def __init__(self, size, dtype, initval=None):
        self.size = size
        if initval is None:
            self._array = np.empty(size, dtype=dtype)
        else:
            self._array = np.full(size, initval, dtype=dtype)
        self.dtype = dtype

    def getblock(self, yoff, xoff, ny, nx):
        return self._array[yoff:yoff+ny, xoff:xoff+nx]

    def setblock(self, yoff, xoff, array):
        (ny, nx) = array.shape
        self._array[yoff:yoff+ny, xoff:xoff+nx] = array
        return

class CompressedBand(object):
    """ CompressedBand is a chunked, blosc-compressed array. """
    CHUNKSET = 1
    CHUNKUNSET = 0

    def __init__(self, size, dtype, chunksize=(256, 256), initval=0):
        """ Initialize a CompressedBand instance.

        Parameters
        ----------
        size : tuple of two ints
            size of band in pixels
        dtype : type
            data type of pixel values
        chunksize : tuple of two ints, optional
            size of compressed chunks, default (256, 256)
        initval : value, optional
            if set, the entire grid is initialized with this value, which should
            be of *dtype*
        """
        assert len(size) == 2
        self.size = size
        self.dtype = dtype
        self._chunksize = chunksize
        self._initval = initval

        self.nchunkrows = int(ceil(float(size[0])/float(chunksize[0])))
        self.nchunkcols = int(ceil(float(size[1])/float(chunksize[1])))
        nchunks = self.nchunkrows * self.nchunkcols

        # Data store
        self._data = [None for i in range(nchunks)]

        # 0 => unset
        # 1 => set
        self.chunkstatus = np.zeros(nchunks, dtype=np.int8)
        return

    def _store(self, array, index):
        self._data[index] = blosc.compress(array.tostring(),
                                           np.dtype(self.dtype).itemsize)
        self.chunkstatus[index] = self.CHUNKSET
        return

    def _retrieve(self, index):
        bytestr = blosc.decompress(self._data[index], as_bytearray=True)
        return np.frombuffer(bytestr, dtype=self.dtype).reshape(self._chunksize)

    def _getchunks(self, yoff, xoff, ny, nx):
        """ Return a generator returning tuples identifying chunks covered by a
        range. The tuples contain (chunk_number, ystart, yend, xstart, xend)
        for each chunk touched by a region defined by corner indices and region
        size. """
        chunksize = self._chunksize
        ystart = yoff // chunksize[0]
        yend = ceil(float(yoff+ny) / chunksize[0])
        xstart = xoff // chunksize[1]
        xend = ceil(float(xoff+nx) / chunksize[1])

        nxchunks = int(ceil(float(self.size[1])/float(chunksize[1])))

        i = ystart
        while i < yend:

            j = xstart

            while j < xend:
                chunk_number = i*nxchunks + j
                chunk_ystart = i*chunksize[0]
                chunk_xstart = j*chunksize[1]
                chunk_yend = min((i+1)*chunksize[0], self.size[0])
                chunk_xend = min((j+1)*chunksize[1], self.size[1])
                yield (chunk_number, chunk_ystart, chunk_yend,
                       chunk_xstart, chunk_xend)
                j += 1

            i+= 1

    def setblock(self, yoff, xoff, array):
        """ Store block of values in *array* starting at offset *yoff*, *xoff*.
        """
        size = array.shape[:2]
        chunksize = self._chunksize

        for i, yst, yen, xst, xen in self._getchunks(yoff, xoff, *size):

            # Get from data store
            if self.chunkstatus[i] == self.CHUNKSET:
                chunkdata = self._retrieve(i)
            else:
                chunkdata = np.full(self._chunksize, self._initval, dtype=self.dtype)

            # Compute region within chunk to place data in
            cy0 = max(0, yoff-yst)
            cy1 = min(chunksize[0], yoff+size[0]-yst)
            cx0 = max(0, xoff-xst)
            cx1 = min(chunksize[1], xoff+size[1]-xst)

            # Compute region to slice from data
            dy0 = max(0, yst-yoff)
            dy1 = min(size[0], yen-yoff)
            dx0 = max(0, xst-xoff)
            dx1 = min(size[1], xen-xoff)

            chunkdata[cy0:cy1, cx0:cx1] = array[dy0:dy1, dx0:dx1]

            # Return to data store
            self._store(chunkdata, i)
        return

    def getblock(self, yoff, xoff, ny, nx):
        """ Retrieve values with dimensions *size*, starting at offset *yoff*,
        *xoff*.
        """
        result = np.empty([ny, nx], self.dtype)
        for i, yst, yen, xst, xen in self._getchunks(yoff, xoff, ny, nx):

            # Compute the bounds in the output
            oy0 = max(0, yst-yoff)
            oy1 = min(ny, yen-yoff)
            ox0 = max(0, xst-xoff)
            ox1 = min(nx, xen-xoff)

            if self.chunkstatus[i] == self.CHUNKUNSET:
                result[oy0:oy1, ox0:ox1] = np.full((oy1-oy0, ox1-ox0),
                                                   self._initval,
                                                   dtype=self.dtype)

            else:
                # Compute the extents from the chunk to retain
                cy0 = max(yoff, yst) - yst
                cy1 = min(yoff+ny, yen) - yst
                cx0 = max(xoff, xst) - xst
                cx1 = min(xoff+nx, xen) - xst

                result[oy0:oy1, ox0:ox1] = self._retrieve(i)[cy0:cy1, cx0:cx1]

        return result




"""
Geographical measurement and analysis

Provides Point, Line, and Polygon classes, and their Multipart equivalents,
with methods for simple measurements such as distance, area, and direction.
"""
from __future__ import division

import math
import itertools
import numbers
import numpy as np
from .decorators import cache_decorator
from ._geojson import GeoJSONOutMixin
from ._shp import ShapefileOutMixin
from .table import Table, Indexer
from .utilities import _reproject, _flatten, _as_nested_lists
from .coordstring import CoordString
from .rtree import RTree
from .quadtree import QuadTree
from . import vectorgeo as _cvectorgeo
from . import dateline as _cdateline
from . import intersection as _cintersection
from . import convexhull as _cconvexhull
from . import contains as _ccontains
from .. import geodesy
from ..crs import Cartesian, CartesianCRS, GeographicalCRS
from ..crs import SphericalEarth
from ..errors import GeometryError, CRSError

class Geometry(object):
    """ Abstract base class for all geometry types """

    def __init__(self, properties=None, crs=Cartesian):
        self.crs = crs
        if isinstance(properties, dict):
            self.properties = properties
        elif properties is not None:
            raise TypeError("properties must be a dictionary")
        else:
            self.properties = {}
        self._cache = {}
        self._geotype = None
        return

class Rotatable(object):

    def rotate(self, thetad, origin=(0, 0)):
        """ Rotate rank 2 geometry.

        Parameters
        ----------
        thetad : float
            degreesof rotation
        origin : tuple of two floats, optional
            pivot for rotation (default (0, 0))
        """
        ct = math.cos(thetad*math.pi / 180.0)
        st = math.sin(thetad*math.pi / 180.0)
        x, y = origin
        M = np.array([[ct, -st, -x*ct + y*st + x],
                      [st, ct, -x*st - y*ct + y]], dtype=np.float64)
        return self.apply_transform(M)

class Point(Geometry, Rotatable, GeoJSONOutMixin, ShapefileOutMixin):
    """ Point object instantiated with:

    Parameters
    ----------
    coords : 2-tuple or 3-tuple
    properties : dict, optional
        geometry specific metadata (default None)
    crs : karta.crs.CRS, optional
        coordinate system for geometry (default Cartesian)
    """
    def __init__(self, coords, properties=None, **kwargs):
        if len(coords) not in (2, 3):
            raise TypeError("Point coordinates must have length 2 or 3")
        super(Point, self).__init__(properties=properties, **kwargs)
        self.vertex = tuple(coords)
        self._geotype = "Point"
        return

    def __getitem__(self, idx):
        return self.vertex[idx]

    def __repr__(self):
        return 'Point({0}, {1})'.format(self.x, self.y)

    def __eq__(self, other):
        try:
            return (tuple(self.vertex) == tuple(other.vertex)) and \
                   (self.properties == other.properties) and \
                   (self.crs == other.crs)
        except AttributeError:
            return False

    def __neq__(self, other):
        return not (self == other)

    def __hash__(self):
        ht = hash(self._geotype)
        hd = hash(self.vertex)
        hc = hash(str(self.crs))
        return ht + (hd << 1) + hd + hc

    def __add__(self, other):
        return Multipoint([self.vertex, other.get_vertex(self.crs)], crs=self.crs)

    @property
    def __geo_interface__(self):
        p = self.properties.copy()
        p["_karta_proj4"] = self.crs.get_proj4()
        return {"type": "Feature",
                "geometry": self.geomdict,
                "properties": p}

    @property
    def geomdict(self):
        return {"type" : "Point", "coordinates" : self.vertex}

    @property
    def x(self):
        return self.vertex[0]

    @property
    def y(self):
        return self.vertex[1]

    @property
    def z(self):
        return self.vertex[2]

    def get_vertex(self, crs=None):
        """ Return the Point vertex as a tuple. """
        if crs is None or crs==self.crs:
            return self.vertex
        else:
            return _reproject((self.x, self.y), self.crs, crs)

    def azimuth(self, other, projected=True):
        """ Returns the compass azimuth from self to other in degrees, measured
        clockwise with north at 0.

        Parameters
        ----------
        other : Point
            second point defining direction
        projected : bool, optional
            If True and self.crs is a ProjectedCRS, return the azimuth in the
            projected cooridnate system. If False, return the geodetic azimuth.
            if self.crs is a GeographicalCRS, result is always geodetic and this
            option has no effect.

        Returns
        -------
        float

        Notes
        -----
        - Return value is NaN if points are coincident
        - If CRS is geographical, returns azimuth as defined by the CRS
          instance.
        """
        az = np.nan
        if projected and not isinstance(self.crs, GeographicalCRS):
            x0, y0 = self.get_vertex()[:2]
            x1, y1 = other.get_vertex(self.crs)[:2]
            if (x0 != x1) or (y0 != y1):
                az = 90.0 - math.atan2(y1-y0, x1-x0)*180.0/math.pi
                az = (az+180) % 360 - 180
        else:
            lon0, lat0 = self.crs.project(self.x, self.y, inverse=True)
            lon1, lat1 = other.crs.project(other.x, other.y, inverse=True)
            if (lon0 != lon1) or (lat0 == lat1):
                az, _, _ = self.crs.inverse(lon0, lat0, lon1, lat1)
        return az

    def apply_transform(self, M):
        """ Apply an affine transform given by matrix *M* to data and return a
        new Point.

        Parameters
        ----------
        M : ndarray
            2x3 or 3x4 affine transformation matrix, representing a
            two-dimensional or a three-dimensional transformation,
            respectively.

        Returns
        -------
        Geometry

        Notes
        -----
        The output coordinates are computed as::

            xnew        x
                  = M * y
            ynew        1

        or::

            xnew        x
            ynew  = M * y
            znew        z
                        1
        """
        if M.shape == (2, 3):
            N = np.zeros((3, 4), dtype=np.float64)
            N[:2,:2] = M[:,:2]
            N[:2,3] = M[:,2]
            N[2, 2] = 1.0
            M = N
        elif M.shape != (3, 4):
            raise ValueError("invalid affine matrix size: {0}".format(M.shape))

        if len(self.vertex) == 2:
            old_vertex = (self.x, self.y, 0)
        else:
            old_vertex = self.vertex

        x = old_vertex[0]*M[0,0] + old_vertex[1]*M[0,1] + old_vertex[2]*M[0,2] + M[0,3]
        y = old_vertex[0]*M[1,0] + old_vertex[1]*M[1,1] + old_vertex[2]*M[1,2] + M[1,3]
        if len(self.vertex) == 2:
            return Point((x, y), properties=self.properties, crs=self.crs)
        else:
            z = old_vertex[0]*M[2,0] + old_vertex[1]*M[2,1] + old_vertex[2]*M[2,2] + M[2,3]
            return Point((x, y, z), properties=self.properties, crs=self.crs)

    def walk(self, distance, azimuth, projected=True):
        """ Returns the point reached when moving in a given direction for
        a given distance from a specified starting location.

        Parameters
        ----------
        distance : float
            distance to shift point
        azimuth: float
            shift azimuth (clockwise with north at 0)
        projected: bool, optional
            If True and self.crs is a ProjectedCRS, return the compute new
            position using the projected cooridnate system. If False, return
            the geodetically correct new position. if self.crs is a
            GeographicalCRS, result is always geodetic and this option has no
            effect.
        """
        if projected and not isinstance(self.crs, GeographicalCRS):
            maz = -(azimuth+90)*math.pi/180
            maz = (90-azimuth) * math.pi/180
            x = self.x + distance*math.cos(maz)
            y = self.y + distance*math.sin(maz)
        else:
            x1, y1 = self.crs.project(self.x, self.y, inverse=True)
            x2, y2, _ = self.crs.forward(x1, y1, azimuth, distance, radians=False)
            x, y = self.crs.project(x2, y2)
        return Point((x, y), properties=self.properties, crs=self.crs)

    def distance(self, other, projected=True):
        """ Returns a distance to another Point. Distances is computed in one
        of three ways:

        1. If *self* is in a geographical coordinate system, *other* is inverse
        projected to geographical coordinates if necessary, and the geodetic
        distance is computed on the ellipsoid of *self*.

        2. If *projected* is `True`, *other* is projected to the same
        coordinate system as *self* and flat-earth distance is computed.

        3. If *projected is `False`, both *self* and *other* are inverse
        projected to geographical coordinates, and the geodetic distance is
        computed on the ellipsoid of *self*.

        If the coordinate system is geographical and a third (z) coordinate
        exists, it is assumed to have the same units as the real-world
        horizontal coordinates (e.g. meters).

        Parameters
        ----------
        other : Point
            point to compute distance to
        projected : bool, optional
            If True and self.crs is a ProjectedCRS, return the flat distance in
            the projected coordinate system. If False, return the ellipsoidal
            distance. if self.crs is a GeographicalCRS, result is always
            ellipsoidal/geodetic and this switch is ignored.

        Returns
        -------
        float

        Notes
        -----
        - If CRS is Geographical, returns distance as computed by the CRS
          instance (e.g. ellipsoidal or spherical).
        """
        if isinstance(self.crs, GeographicalCRS):
            lon0, lat0 = self.x, self.y
            lon1, lat1 = other.crs.project(other.x, other.y, inverse=True)
            _, _, dist = self.crs.inverse(lon0, lat0, lon1, lat1)
        elif projected:
            x0, y0 = self.vertex[:2]
            x1, y1 = other.get_vertex(self.crs)[:2]
            dist = math.sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0))
        else:
            lon0, lat0 = self.crs.project(self.x, self.y, inverse=True)
            lon1, lat1 = other.crs.project(other.x, other.y, inverse=True)
            _, _, dist = self.crs.inverse(lon0, lat0, lon1, lat1)

        if 3 == len(self.vertex) == len(other.vertex):
            dz = self.z - other.z
            dist = math.sqrt(dist*dist + dz*dz)
        return dist

    def shift(self, shift_vector):
        """ Shift point in space.

        Parameters
        ----------
        shift_vector : iterable
            vector with length equal to Geometry rank defining the shift
        inplace : bool
            whether shift should be in place (default False)
        """
        if len(shift_vector) != 2:
            raise ValueError("shift vector must be of the form (xshift, yshift)")
        if len(self.vertex) == 2:
            vertex = (self.x+shift_vector[0], self.y+shift_vector[1])
        else:
            vertex = (self.x+shift_vector[0], self.y+shift_vector[1], self.z)
        return Point(vertex, properties=self.properties, crs=self.crs)

class MultiVertexBase(Geometry):

    def __init__(self, vertices, ring=False, **kwargs):
        super(MultiVertexBase, self).__init__(**kwargs)

        if hasattr(vertices, "__next__"):
            vertices = list(vertices)

        if len(vertices) == 0:
            self.vertices = CoordString([], ring=ring)
        elif not isinstance(vertices[0], Point):
            self.vertices = CoordString(_flatten(vertices), ring=ring)
        else:
            self.vertices = CoordString([point.vertex for point in vertices], ring=ring)
            if "crs" not in kwargs:
                self.crs = vertices[0].crs

        if self.vertices.rank not in (2, 3):
            raise TypeError("Coordinates must be a sequence of sequences")
        return

    def __eq__(self, other):
        try:
            return (self._geotype == other._geotype) and \
                   (len(self.vertices) == len(other.vertices)) and \
                   np.all(np.equal(self.vertices, other.vertices)) and \
                   (self.properties == other.properties) and \
                   (self.crs == other.crs)
        except (AttributeError, TypeError, ValueError):
            return False

    def __neq__(self, other):
        return ~(self == other)

    def __hash__(self):
        ht = hash(self._geotype)
        hd = hash(self.vertices.asarray().tostring())
        hc = hash(str(self.crs))
        return ht + (hd << 1) + hd + hc

    def __getitem__(self, key):
        if isinstance(key, (int, np.int64)):
            return Point(self.vertices[key], properties=self.properties, crs=self.crs)
        elif isinstance(key, slice):
            start, stop, stride = key.indices(len(self.vertices))
            verts = self.vertices.slice(start, stop, stride)
            return type(self)(verts, properties=self.properties, crs=self.crs)
        else:
            raise KeyError('index must be integer or slice object')

    def __iter__(self):
        return (self[i] for i in range(len(self)))

    def __len__(self):
        return len(self.vertices)

class MultiVertexMixin(object):

    def get_coordinate_lists(self, crs=None):
        """ Return horizontal coordinate lists.

        Parameters
        ----------
        crs : karta.CRS, optional
            coordinate system of output vertices
        """
        x, y = self.vertices.vectors()[:2]
        if crs is not None and (crs != self.crs):
            x, y = _reproject((x,y), self.crs, crs)
        return x, y

    @property
    def coordinates(self):
        return self.get_coordinate_lists()

    def get_vertices(self, crs=None, drop_z=False):
        """ Return vertices as an array.

        Parameters
        ----------
        crs : karta.CRS, optional
            coordinate system of output vertices
        drop_z : bool, optional
            whether to discard third dimension for rank 3 geometries
        """
        if (crs is None) or (crs is self.crs):
            if not drop_z:
                return self.vertices.asarray()
            else:
                return self.vertices.asarray()[:,:2]
        else:
            v = self.vertices.vectors(drop_z=drop_z)
            vt = _reproject(v, self.crs, crs)
            return np.vstack(vt).T

    @property
    def bbox(self):
        return self.get_bbox()

    @cache_decorator("bbox")
    def get_bbox(self, crs=None):
        """ Return the bounding box.

        Parameters
        ----------
        crs : karta.CRS
            coordinate system of output bounding box

        Returns
        -------
        tuple
            (xmin, ymin, xmax, ymax)

        Notes
        -----
        - If CRS is Geographical, returns bbox computed using a spherical
          approximation.
        """
        if crs is not None and (crs != self.crs):
            cs = CoordString(list(zip(
                *self.crs.transform(crs, *self.vertices.vectors(drop_z=True)))))
        else:
            cs = self.vertices
            crs = self.crs

        if isinstance(crs, GeographicalCRS):
            return _cdateline.bbox(cs)
        else:
            return _cvectorgeo.bbox(cs)

    @property
    def extent(self):
        return self.get_extent()

    @cache_decorator("extent")
    def get_extent(self, crs=None):
        """ Calculate geometry extent.

        Parameters
        ----------
        crs : karta.CRS
            coordinate system of output

        Returns
        -------
        tuple
            (xmin, xmax, ymin, ymax)
        """
        bb = self.get_bbox(crs=crs)
        return bb[0], bb[2], bb[1], bb[3]

    def _bbox_overlap(self, other):
        """ Whether bounding box overlaps with that of another Geometry. """
        reg0 = self.bbox
        reg1 = other.bbox
        return (reg0[0] <= reg1[2] and reg1[0] <= reg0[2] and
                reg0[1] <= reg1[3] and reg1[1] <= reg0[3])

    def apply_transform(self, M):
        """ Apply an affine transform given by matrix *M* to data and return a
        new geometry.

        Parameters
        ----------
        M : ndarray
            2x3 or 3x4 affine transformation matrix, representing a
            two-dimensional or a three-dimensional transformation,
            respectively.

        Returns
        -------
        Geometry

        Notes
        -----
        The output coordinates are computed as::

            xnew        x
                  = M * y
            ynew        1

        or::

            xnew        x
            ynew  = M * y
            znew        z
                        1
        """
        if len(self.vertices) == 0:
            raise ValueError("cannot transform zero length geometry")

        if M.shape == (2, 3):
            N = np.zeros((3, 4), dtype=np.float64)
            N[:2,:2] = M[:,:2]
            N[:2,3] = M[:,2]
            N[2, 2] = 1.0
            M = N
        elif M.shape != (3, 4):
            raise ValueError("invalid affine matrix size: {0}".format(M.shape))

        xy = self.get_vertices()
        if xy.shape[1] == 2:
            rank = 2
            xy = np.hstack([xy, np.zeros((len(xy), 1), dtype=np.float64)])
        else:
            rank = 3
        xy = np.hstack([xy, np.ones((len(xy), 1), dtype=np.float64)])
        new_xy = np.dot(M, xy.T).T
        if rank == 2:
            new_xy = new_xy[:,:2]

        if hasattr(self, "data"):
            return type(self)(new_xy, properties=self.properties, data=self.data, crs=self.crs)
        else:
            return type(self)(new_xy, properties=self.properties, crs=self.crs)

    def shift(self, shift_vector):
        """ Shift geometry in space.

        Parameters
        ----------
        shift_vector : iterable
            vector with length equal to Geometry rank defining the shift
        """
        if len(shift_vector) != 2:
            raise ValueError("shift vector must be of the form (xshift, yshift)")
        trans_mat = np.array([[1, 0, shift_vector[0]],
                              [0, 1, shift_vector[1]]], dtype=np.float64)
        return self.apply_transform(trans_mat)

    def _subset(self, idxs):
        """ Return a subset defined by indices. """
        vertices = [self.vertices[i] for i in idxs]
        if hasattr(self, "data"):
            data = Table(data=[self.data._data[i] for i in idxs], fields=self.data.fields)
            return type(self)(vertices, properties=self.properties, data=data, crs=self.crs)
        else:
            return type(self)(vertices, properties=self.properties, crs=self.crs)

    def flat_distances_to(self, pt):
        """ Return the "flat Earth" distance from each vertex to a point. """
        A = self.vertices.asarray()
        P = np.tile(np.array(pt.vertex), (A.shape[0], 1))
        d = np.sqrt(np.sum((A-P)**2, 1))
        return d

    def distances_to(self, pt):
        """ Return the distance from each vertex to a point. """
        d = [pt.distance(a) for a in self]
        return np.array(d)

    def nearest_vertex_to(self, point):
        """ Returns the index of the vertex that is nearest to a point. If two
        points are equidistant, only one will be returned.

        Parameters
        ----------
        point : Point
            target point

        Returns
        -------
        int
        """
        distances = self.distances_to(point)
        idx = np.argmin(distances)
        return idx

    def any_within_poly(self, poly):
        """ Return whether any vertices are inside *poly* """
        for pt in self:
            if poly.contains(pt):
                return True
        return False

    def convex_hull(self):
        """ Return a Polygon representing the convex hull.

        Notes
        -----
        - If CRS is Geographical, returns hull computed using a spherical
          approximation. Failure may occur if the vertices are not in euclidian
          position.
        """
        if isinstance(self.crs, GeographicalCRS):
            indices = _cconvexhull.convexhull_sph(self.vertices)
        else:
            indices = _cconvexhull.convexhull(self.vertices)
        return Polygon([self.vertices[i] for i in indices], crs=self.crs)


class ConnectedMultiVertexMixin(MultiVertexMixin):

    @cache_decorator("bbox")
    def get_bbox(self, crs=None):
        """ Dateline-aware get_bbox for geometries consisting of connected
        vertices.

        Parameters
        ----------
        crs : karta.CRS
            coordinate system of output bounding box

        Returns
        -------
        tuple
             (xmin, ymin, xmax, ymax)
        """
        if crs is not None and (crs != self.crs):
            cs = CoordString(list(zip(
                *self.crs.transform(crs, *self.vertices.vectors(drop_z=True)))))
        else:
            cs = self.vertices
            crs = self.crs

        if isinstance(crs, GeographicalCRS):
            bbox = _cdateline.bbox(cs)
        else:
            bbox = super(ConnectedMultiVertexMixin, self).get_bbox(crs=crs)
        return bbox

    @property
    @cache_decorator("length")
    def length(self):
        """ Returns the length of the line/boundary.

        Notes
        -----
        - If CRS is Geographical, returns length computed using distance
          provided by the CRS instance.
        """
        if isinstance(self.crs, GeographicalCRS):
            lon, lat = self.get_vertices()[0][:2]
            d = 0.0
            for xy in self.get_vertices()[1:]:
                d += self.crs.inverse(lon, lat, xy[0], xy[1])[2]
                lon = xy[0]
                lat = xy[1]
            return d
        else:
            return _cvectorgeo.length(self.vertices)

    @property
    def segments(self):
        """ Returns an generator of adjacent line segments. """
        return (self._subset((i,i+1)) for i in range(len(self)-1))

    @property
    def segment_tuples(self):
        """ Returns an generator of adjacent line segments as coordinate tuples. """
        return ((self.vertices[i], self.vertices[i+1])
                for i in range(len(self.vertices)-1))

    def intersects(self, other):
        """ Return whether an intersection exists with another geometry.

        Parameters
        ----------
        other : Geometry
            another geometry with multiple connected vertices

        Notes
        -----
        - If CRS is Geographical, uses a spherical approximation.
        """
        if _cintersection.bboxes_overlap(self.bbox, other.bbox):
            if isinstance(self.crs, GeographicalCRS):
                return _cintersection.intersects_sph(self.vertices, other.vertices)
            else:
                return _cintersection.intersects(self.vertices, other.vertices)

    def intersections(self, other, keep_duplicates=False):
        """ Return the intersections with another geometry as a Multipoint.

        Parameters
        ----------
        other : Geometry
            another geometry with multipl connected vertices
        keep_duplicates : bool, optional
            whether to retain duplicate intersections [default False]

        Notes
        -----
        - If CRS is Geographical, uses a spherical approximation.
        """
        if isinstance(self.crs, CartesianCRS):
            interx = _cintersection.all_intersections(self.vertices, other.vertices)
            if not keep_duplicates:
                interx = list(set(interx))
            return Multipoint(interx, crs=self.crs)
        else:
            interx = (geodesy.intersection_spherical(a, b)
                            for a in self.segment_tuples
                            for b in other.segment_tuples)
            if not keep_duplicates:
                interx = list(set(interx))
            return Multipoint(interx, crs=self.crs)

    def _nearest_to_point(self, point):
        """ Return a tuple of the shortest distance on the geometry boundary to
        a point, and the vertex at that location.

        If necessary, project coordinates to the local coordinate system.

        Parameters
        ----------
        point : Point

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        ptvertex = point.get_vertex(crs=self.crs)
        segments = zip(self.vertices.slice(0, -1), self.vertices.slice(1, 0))

        if isinstance(self.crs, CartesianCRS):
            func = _cvectorgeo.pt_nearest_planar
            def func(seg):
                return _cvectorgeo.pt_nearest_planar(ptvertex[0], ptvertex[1],
                                    seg[0][0], seg[0][1], seg[1][0], seg[1][1])
        else:
            fwd = self.crs.forward
            inv = self.crs.inverse
            def func(seg):
                return _cvectorgeo.pt_nearest_proj(fwd, inv, ptvertex,
                                                   seg[0], seg[1], tol=0.01)

        point_dist = map(func, segments)
        min_point = None
        min_dist = -1.0
        for i, (point, dist) in enumerate(point_dist):
            if dist < min_dist or (i == 0):
                min_point = point
                min_dist = dist

        return min_dist, min_point

    def shortest_distance_to(self, pt):
        """ Return the shortest distance from any position on the geometry
        boundary to a point.

        Parameters
        ----------
        point : Point

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        return self._nearest_to_point(pt)[0]

    def nearest_on_boundary(self, point):
        """ Returns the position on the geometry boundary that is nearest to
        a point. If two points are equidistant, only one will be returned.

        Parameters
        ----------
        point : Point

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        _, minpt = self._nearest_to_point(point)
        return Point(minpt, crs=self.crs)

    def within_distance(self, point, distance):
        """ Test whether a point is within *distance* geometry.

        Parameters
        ----------
        point : Point
        distance : float

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        return all(distance >= seg.shortest_distance_to(point)
                    for seg in self.segments)

    def crosses_dateline(self):
        """ Return a boolean that indicates whether any segment crosses the
        dateline.
        """
        if not isinstance(self.crs, GeographicalCRS):
            raise CRSError("Dateline detection only defined for geographical "
                           "coordinates")

        def _seg_crosses_dateline(seg):
            a, b = seg[0], seg[1]
            return (_sign(a.x) != _sign(b.x)) and (abs(a.x-b.x) > 180.0)

        return any(_seg_crosses_dateline(seg) for seg in self.segments)

class Line(MultiVertexBase, ConnectedMultiVertexMixin, Rotatable, GeoJSONOutMixin, ShapefileOutMixin):
    """ Line composed of connected vertices.

    Parameters
    ----------
    coords : list of 2-tuples or 3-tuples
        vertex coordinates
    properties : dict, optional
        geometry specific metadata
    crs : karta.CRS, optional
        (default Cartesian)
    """
    def __init__(self, vertices, **kwargs):
        """ Partial init function that creates a metadata attribute.
        """
        super(Line, self).__init__(vertices, **kwargs)
        self._geotype = "Line"
        return

    def __add__(self, other):
        return Multiline([self.vertices, other.get_vertices(self.crs)],
                         crs=self.crs)

    @property
    def __geo_interface__(self):
        p = self.properties.copy()
        p["_karta_proj4"] = self.crs.get_proj4()
        return {"type": "Feature",
                "geometry": self.geomdict,
                "properties": p}

    @property
    def geomdict(self):
        return {"type" : "LineString",
                "bbox" : self.bbox,
                "coordinates" : _as_nested_lists(self.vertices)}

    def extend(self, other):
        """ Combine two lines, provided that that the data formats are similar.
        """
        if len(self.vertices[0]) != len(other.vertices[0]):
            raise ValueError("Rank mismatch ({0} != "
                    "{1})".format(self.vertices.shape[1],
                                  other.vertices.shape[1]))
        if self._geotype != other._geotype:
            raise TypeError("Geometry mismatch ({0} != "
                    "{1})".format(self._geotype, other._geotype))

        self.vertices = np.vstack([self.vertices, other.vertices])
        self._cache = {}
        return self

    def cumulength(self):
        """ Returns the cumulative length by vertex.

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        d = [0.0]
        pta = self[0]
        for ptb in self[1:]:
            d_ = pta.distance(ptb)
            d.append(d_ + d[-1])
            pta = ptb
        return d

    def to_points(self, dx):
        """ Return equally spaced Point instances along line.

        Parameters
        ----------
        dx : float
            spacing of points
        """
        remainder = 0
        pt0 = self[0]
        vertices = [pt0.get_vertex()]

        for seg in self.segments:
            pos = 0
            az = seg[0].azimuth(seg[1])

            while pos < seg.length:
                distance_to_endpt = pt0.distance(seg[1])
                if distance_to_endpt >= dx:
                    pt1 = pt0.walk(dx - remainder, az)
                    pos += dx - remainder
                    vertices.append(pt1.get_vertex())
                    remainder = 0
                    pt0 = pt1
                else:
                    remainder = distance_to_endpt
                    pos = seg.length
                    pt0 = seg[1]
        return Multipoint(vertices, crs=self.crs)

    def to_npoints(self, n):
        """ Return *n* equally spaced Point instances along line.

        Parameters
        ----------
        n : int
            number of points to return
        """
        segments = self.segments
        Ltotal = self.cumulength()[-1]
        step = Ltotal / float(n-1)
        step_remaining = step

        vertices = [self[0].get_vertex()]
        x = 0.0
        pos = self[0]
        seg = next(segments)
        seg_remaining = seg.displacement()

        while x < Ltotal-1e-8:
            direction = seg[0].azimuth(seg[1])

            if step_remaining <= seg_remaining:
                pos = pos.walk(step_remaining, direction)
                x += step_remaining
                seg_remaining -= step_remaining
                step_remaining = step
                vertices.append(pos.get_vertex())
                seg.vertices[0] = np.array(pos.vertex, dtype=np.float64)

            else:
                pos = seg[1]
                x += seg_remaining
                step_remaining -= seg_remaining

                seg = next(segments, seg)
                seg_remaining = seg.displacement()

        if len(vertices) == n-1:
            vertices.append(seg[-1].get_vertex())
        return Multipoint(vertices, crs=self.crs)

    def displacement(self):
        """ Returns the distance between the first and last vertex.

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        return self[0].distance(self[-1])

    def to_polygon(self):
        """ Returns a polygon. """
        return Polygon(self.vertices, properties=self.properties, crs=self.crs)

class Polygon(MultiVertexBase, Rotatable, ConnectedMultiVertexMixin, GeoJSONOutMixin, ShapefileOutMixin):
    """ Polygon, composed of a closed sequence of vertices.

    Parameters
    ----------
    coords : list of 2-tuples or 3-tuples
        vertex coordinates
    subs : list of Polygon instances, optional
        sub-polygons [default None]
    properties : dict, optional
        geometry specific metadata
    crs : karta.CRS, optional
        (default Cartesian)
    """
    def __init__(self, vertices, subs=None, **kwargs):
        """ Partial init function that creates a metadata attribute.
        """
        super(Polygon, self).__init__(vertices, ring=True, **kwargs)
        self._geotype = "Polygon"
        if subs is not None:
            self.subs = list(subs)
        else:
            self.subs = []
        return

    def __getitem__(self, key):
        if isinstance(key, slice):
            start, stop, stride = key.indices(len(self.vertices))
            if len(self) != ((stop - start) // stride):
                return Line(self.vertices.slice(start, stop, stride),
                            properties=self.properties,
                            crs=self.crs)
        return super(Polygon, self).__getitem__(key)

    def __add__(self, other):
        return Multipolygon([[self.vertices], [other.get_vertices(self.crs)]],
                            crs=self.crs)

    @property
    def __geo_interface__(self):
        p = self.properties.copy()
        p["_karta_proj4"] = self.crs.get_proj4()
        return {"type": "Feature",
                "geometry": self.geomdict,
                "properties": p}

    @property
    def vertices_ring(self):
        # inefficient implementation
        vertices = [xy for xy in self.vertices]
        vertices.append(vertices[0])
        return CoordString(vertices)

    @property
    def geomdict(self):
        coords = [_as_nested_lists(self.vertices_ring)]
        for geom in self.subs:
            coords.append(_as_nested_lists(geom.vertices_ring))
        return {"type" : "Polygon",
                "bbox" : self.bbox,
                "coordinates" : coords}

    def _subset(self, idxs):
        """ Return a subset defined by index in *idxs*. """
        vertices = [self.vertices[i] for i in idxs]
        subset = Line(vertices, properties=self.properties, crs=self.crs)
        return subset

    def isclockwise(self):
        """ Return whether polygon winds clockwise around its interior. """
        s = sum((seg[1][0] - seg[0][0]) * (seg[1][1] + seg[0][1])
                        for seg in self.segment_tuples)
        return s > 0

    def ispolar(self, pole=None):
        """ Return True if polygon contains one pole. If the polygon contains
        neither or both poles, returns False.

        Parameters
        ----------
        pole : Point, optional
            (default point on a sphere at 0 longitude, 90 latitude)
        """

        if not isinstance(self.crs, GeographicalCRS):
            raise CRSError("ispolar defined only for geographical CRS")

        if pole is None:
            pole = Point((0, 90), crs=SphericalEarth)

        lon0 = geodesy.reduce_deg(self[-1].vertex[0])
        sum_angle = 0.0
        for vertex in self.vertices:
            lon1 = geodesy.reduce_deg(vertex[0])
            if _cdateline.crosses_dateline(lon0, lon1):
                sum_angle += 360.0 + lon1 - lon0
            else:
                sum_angle += lon1 - lon0
            lon0 = lon1

        return True if abs(sum_angle) > 1e-4 else False

    @property
    def segments(self):
        """ Returns a generator of adjacent line segments.
        """
        L = len(self.vertices)
        return itertools.chain((self._subset((i,i+1)) for i in range(len(self)-1)),
                               (self._subset((L-1,0)),))

    @property
    def segment_tuples(self):
        """ Returns a generator of adjacent line segments as coordinate
        tuples. """
        return ((self.vertices[i-1], self.vertices[i])
                for i in range(len(self.vertices)))

    @property
    def length(self):
        raise AttributeError("%s instance has no attribute 'length'" % type(self))

    @property
    def perimeter(self):
        """ Return the perimeter of the polygon. If there are sub-polygons,
        their perimeters are added recursively.

        Notes
        -----
        - If CRS is Geographical, uses distance defined by the CRS instance.
        """
        return sum(seg.length for seg in self.segments) + \
                sum([p.perimeter for p in self.subs])

    @property
    def area(self):
        """ Return the two-dimensional area of the polygon, excluding
        sub-polygons.

        Notes
        -----
        - If CRS is Geographical, uses either a spherical or an ellipsoidal
          calculation.
        """
        if isinstance(self.crs, GeographicalCRS):
            major_axis = self.crs.ellipsoid.a
            minor_axis = self.crs.ellipsoid.b

            area = 0.0
            if major_axis == minor_axis:    # Sphere
                for seg in self.segment_tuples:
                    x1, y1 = seg[0]
                    x2, y2 = seg[1]
                    area += geodesy.spherical_area(major_axis, x1, y1, x2, y2)

            else:
                for seg in self.segment_tuples:
                    x1, y1 = seg[0]
                    x2, y2 = seg[1]
                    area += geodesy.ellipsoidal_area(major_axis, minor_axis,
                                                     x1, y1, x2, y2)

        else:
            # Cartesian coordinate systems
            x, y = self.coordinates
            x0 = np.min(x)
            area = (0.5*(x[0] + x[-1]) - x0) * (y[0] - y[-1])
            area += sum((0.5*(x[i+1]+x[i]) - x0) * (y[i+1] - y[i]) for i in range(len(x)-1))
        return abs(area) - sum(sub.area for sub in self.subs)

    @property
    def centroid(self):
        """ Return Polygon centroid as a Point, ignoring sub-polygons. """
        x, y = self.coordinates
        A = 0.5 * sum(x[i]*y[i+1] - x[i+1]*y[i] for i in range(-1, len(self)-1))
        cx = sum((x[i] + x[i+1]) * (x[i]*y[i+1] - x[i+1]*y[i])
                    for i in range(-1, len(self)-1)) / (6*A)
        cy = sum((y[i] + y[i+1]) * (x[i]*y[i+1] - x[i+1]*y[i])
                    for i in range(-1, len(self)-1)) / (6*A)
        return Point((cx, cy), properties=self.properties, crs=self.crs)

    def contains(self, point):
        """ Returns True if point is inside or on the boundary of the polygon,
        and False otherwise. Uses a crossing number scheme.

        Notes
        -----
        - When the polygon is polar in a geographical coordinate system, a less
          efficient algorithm is used. For better performance, consider
          projecting to an appropriate coordinate system such as NSIDCNorth or
          NSIDCSouth beforehand.
        - Otherwise, a planar algorithm is used.
        """
        x, y = point.get_vertex(crs=self.crs)[:2]
        if isinstance(self.crs, GeographicalCRS) and self.ispolar():
            return _ccontains.contains_proj(x, y, self.vertices, self.crs) \
                    and not any(p.contains(point) for p in self.subs)
        else:
            return _ccontains.contains(x, y, self.vertices) \
                    and not any(p.contains(point) for p in self.subs)

    def to_line(self):
        """ Returns a self-closing polyline. Discards sub-polygons. """
        v = self.vertices + self.vertices[0]
        return Line(v, properties=self.properties, crs=self.crs)

class Multipart(Geometry):
    """ Base for objects consisting of multiple singular types. """

    def __init__(self, vertices, data=None, **kwargs):
        super(Multipart, self).__init__(**kwargs)
        if data is None or len(data) == 0:
            self.data = Table(size=len(self.vertices))
        elif isinstance(data, Table):
            self.data = data
        else:
            for k, v in data.items():
                if len(v) != len(vertices):
                    raise ValueError("length of `data` member '{k}' ({n}) not "
                                     "equal to length of vertices ({m})".format(
                                         k=k, n=len(v), m=len(vertices)))
            self.data = Table(data)
        return

    @property
    def d(self):
        return Indexer(self.data)

    def __eq__(self, other):
        try:
            return (self._geotype == other._geotype) and \
                   (len(self.vertices) == len(other.vertices)) and \
                   np.all(np.equal(self.vertices, other.vertices)) and \
                   (self.data == other.data) and \
                   (self.properties == other.properties) and \
                   (self.crs == other.crs)
        except (AttributeError, TypeError):
            return False

    def __neq__(self, other):
        return ~(self == other)

    def __hash__(self):
        ht = hash(self._geotype)
        hd = hash(self.vertices.asarray().tostring())
        hc = hash(str(self.crs))
        return ht + (hd << 1) + hd + hc

    def __contains__(self, other):
        if other in (part for part in self):
            return True
        else:
            return False

    def __len__(self):
        return len(self.vertices)


class Multipoint(Multipart, Rotatable, MultiVertexMixin, GeoJSONOutMixin, ShapefileOutMixin):
    """ Point cloud with associated attributes.

    Parameters
    ----------
    coords : list
        list of 2-tuples or 3-tuples defining vertices
    data : list, dict, Table object, or None
        point-specific data [default None]
    properties : dict or None
        geometry specific data [default None]
    crs : karta.crs.CRS subclass
        [default Cartesian]
    """

    def __init__(self, vertices, build_index=True, **kwargs):
        if hasattr(vertices, "__next__"):
            vertices = list(vertices)

        if len(vertices) == 0:
            self.vertices = CoordString([])
        elif not isinstance(vertices[0], Point):
            self.vertices = CoordString(vertices)
        else:
            self.vertices = CoordString([point.vertex for point in vertices])
            kwargs.setdefault("crs", vertices[0].crs)

        super(Multipoint, self).__init__(vertices, **kwargs)
        if build_index:
            self.quadtree = QuadTree(self.vertices)
        self._geotype = "Multipoint"
        return

    def __getitem__(self, key):
        if isinstance(key, numbers.Integral):
            p = self.d[key]
            p.update(self.properties)
            return Point(self.vertices[key], properties=p, crs=self.crs)
        elif isinstance(key, slice):
            start, stop, stride = key.indices(len(self.vertices))
            return Multipoint(self.vertices.slice(start, stop, stride),
                              properties=self.properties,
                              data=self.d[key], crs=self.crs)
        else:
            raise KeyError(type(key))

    def __setitem__(self, key, value):
        if isinstance(key, numbers.Integral):
            if hasattr(value, "vertex"):
                self.vertices[key] = np.array(value.get_vertex(self.crs), dtype=np.float64)
                row = []
                for field in self.data.fields:
                    row.append(value.properties.get(field, None))
                self.data[key] = tuple(row)
            else:
                if len(value) == len(self.vertices[0]):
                    verts = self.vertices.asarray()
                    verts[key] = value
                    self.vertices = CoordString(verts)
                self.data[key] = (None for _ in self.data.fields)

    @property
    def geomdict(self):
        return {"type" : "MultiPoint",
                "bbox" : self.bbox,
                "coordinates" : _as_nested_lists(self.vertices)}

    @property
    def __geo_interface__(self):
        p = self.properties.copy()
        p["_karta_proj4"] = self.crs.get_proj4()
        return {"type": "Feature",
                "geometry": self.geomdict,
                "properties": p}

    def within_radius(self, point, radius):
        """ Return subset of Multipoint within a radius. Items on the border
        are excluded.

        Parameters
        ----------
        point : Point
            point to to center filter at
        radius : float
            maximum distance from *point*

        Returns
        -------
        Multipoint
        """
        if hasattr(self, "quadtree"):
            candidate_indices = self.quadtree.search_within(point.x-radius,
                                                            point.y-radius,
                                                            point.x+radius,
                                                            point.y+radius)
            confirmed_indices = []
            for i in candidate_indices:
                if point.distance(self[i]) < radius:
                    confirmed_indices.append(i)
            confirmed_indices.sort()
        else:
            confirmed_indices = [i for i,d in enumerate(self.distances_to(point))
                                   if d < radius]
        return self._subset(confirmed_indices)

    def within_bbox(self, bbox):
        """ Return Multipoint subset that is within a square bounding box
        given by (xmin, xymin, xmax, ymax).
        """
        if hasattr(self, "quadtree"):
            indices = self.quadtree.search_within(*bbox)
            indices.sort()
        else:
            indices = [i for (i, pt) in enumerate(self)
                         if (bbox[0] < pt.x < bbox[2]) and (bbox[1] < pt.y < bbox[3])]
        return self._subset(indices)

    def within_polygon(self, poly):
        """ Return Multipoint subset that is within a polygon.
        """
        if hasattr(self, "quadtree"):
            bbox = poly.get_bbox(crs=self.crs)
            candidate_indices = self.quadtree.search_within(*bbox)
            confirmed_indices = []
            for i in candidate_indices:
                if poly.contains(self[i]):
                    confirmed_indices.append(i)
            confirmed_indices.sort()
        else:
            confirmed_indices = [i for (i, point) in enumerate(self)
                                   if poly.contains(point)]
        return self._subset(confirmed_indices)

class MultiVertexMultipartMixin(object):
    """ Mix-in class for multipart classes for which it is reasonable to ask
    whether member geometries are within or touching a multi-vertex geometry.

    E.g.
    - "which members touch this Line/Polygon?"
    - "which members are contained by this Polygon?"
    """
    @cache_decorator("bbox")
    def get_bbox(self, crs=None):
        bbs = [part.get_bbox(crs=crs) for part in self]
        xmin = min([bb[0] for bb in bbs])
        ymin = min([bb[1] for bb in bbs])
        xmax = max([bb[2] for bb in bbs])
        ymax = max([bb[3] for bb in bbs])
        return (xmin, ymin, xmax, ymax)

    @cache_decorator("extent")
    def get_extent(self, crs=None):
        bb = self.get_bbox(crs=crs)
        return bb[0], bb[2], bb[1], bb[3]

    @property
    def bbox(self):
        return self.get_bbox()

    @property
    def extent(self):
        return self.get_extent()

    def apply_transform(self, M):
        """ Apply an affine transform given by matrix *M* to data and return a
        new Point.

        Parameters
        ----------
        M : ndarray
            2x3 or 3x4 affine transformation matrix, representing a
            two-dimensional or a three-dimensional transformation,
            respectively.

        Returns
        -------
        Geometry

        Notes
        -----
        The output coordinates are computed as::

            xnew        x
                  = M * y
            ynew        1

        or::

            xnew        x
            ynew  = M * y
            znew        z
                        1
        """
        if M.shape == (2, 3):
            N = np.zeros((3, 4), dtype=np.float64)
            N[:2,:2] = M[:,:2]
            N[:2,3] = M[:,2]
            N[2, 2] = 1.0
            M = N
        elif M.shape != (3, 4):
            raise ValueError("invalid affine matrix size: {0}".format(M.shape))
        parts = []
        for part in self:
            parts.append(part.apply_transform(M))
        return type(self)(parts, properties=self.properties, data=self.data, crs=self.crs)

    def within_bbox(self, bbox, max_results=-1):
        """ Return Multipart geometry representing member geometries that are
        contained by a bounding box.

        Parameters
        ----------
        bbox : tuple
            (xmin, ymin, xmax, ymax)
        """
        indices = self.rtree.search_within(bbox, max_results=max_results)
        return type(self)([self[i] for i in indices])

    def touching_bbox(self, bbox, max_results=-1):
        """ Return Multipart geometry representing member geometries that touch
        a bounding box.

        Parameters
        ----------
        bbox : tuple
            (xmin, ymin, xmax, ymax)
        """
        indices = self.rtree.search_overlapping(bbox, max_results=max_results)
        return type(self)([self[i] for i in indices])

    def touching(self, geom):
        """ Return a Multipart geometry representing member geometries that
        touch a Line or Polygon.

        Touching is defined as intersecting a Line or Polygon boundary, or
        being contained within a Polygon.

        Parameters
        ----------
        geom : Line or Polygon

        Returns
        -------
        """
        indices = self.rtree.search_overlapping(geom.get_bbox(self.crs))
        results = []
        if isinstance(geom, Line):
            for i in indices:
                test_geom = self[i]
                if geom.intersects(test_geom):
                    results.append(test_geom)
        elif isinstance(geom, Polygon):
            for i in indices:
                test_geom = self[i]
                pt = test_geom[0]
                if geom.contains(pt) or geom.intersects(test_geom):
                    results.append(test_geom)
        else:
            raise TypeError("argument must be Line or Polygon")
        return type(self)(results)

    def within(self, geom):
        """ Return a Multipart geometry representing member geometries
        contained within a Polygon.

        Parameters
        ----------
        geom : Polygon
        """
        if not isinstance(geom, Polygon):
            raise TypeError("argument must be Polygon")
        indices = self.rtree.search_overlapping(geom.bbox)
        results = []
        for i in indices:
            test_geom = self[i]
            pt = test_geom[0]
            if geom.contains(pt) and not geom.intersects(test_geom):
                results.append(test_geom)
        return type(self)(results)

class Multiline(Multipart, MultiVertexMultipartMixin, GeoJSONOutMixin,
                ShapefileOutMixin):
    """ Collection of lines with associated attributes.

    Parameters
    ----------
    coords : list
        list of lists of 2-tuples or 3-tuples defining lines
    data : list, dict, Table object, or None
        point-specific data [default None]
    properties : dict or None
        geometry specific data [default None]
    crs : karta.crs.CRS subclass
        [default Cartesian]
    """

    def __init__(self, vertices, build_index=True, **kwargs):
        if len(vertices) == 0:
            self.vertices = []
        elif isinstance(vertices[0], Line):
            self.vertices = [line.vertices for line in vertices]
            kwargs.setdefault("crs", vertices[0].crs)
        else:
            self.vertices = [CoordString(part) for part in vertices]
        super(Multiline, self).__init__(vertices, **kwargs)
        if build_index:
            self.rtree = RTree(self.vertices)
        self._geotype = "Multiline"
        return

    def __getitem__(self, key):
        if isinstance(key, numbers.Integral):
            properties = self.d[key]
            properties.update(self.properties)
            return Line(self.vertices[key], properties=properties, crs=self.crs)
        elif isinstance(key, slice):
            return Multiline(self.vertices[key], properties=self.properties,
                             data=self.d[key], crs=self.crs)
        else:
            raise KeyError(type(key))

    @property
    def geomdict(self):
        return {"type" : "MultiLineString",
                "bbox" : self.bbox,
                "coordinates" : _as_nested_lists(self.vertices)}

    @property
    def __geo_interface__(self):
        p = dict(_karta_proj4=self.crs.get_proj4())
        return {"type": "Feature",
                "geometry": self.geomdict,
                "properties": p}

    def get_vertices(self, crs=None):
        """ Return vertices as a list of arrays.

        Parameters
        ----------
        crs : karta.CRS, optional
            coordinate system of output vertices
        """
        if crs is None or (crs == self.crs):
            return [v.asarray() for v in self.vertices]
        else:
            vertices = []
            for line in self.vertices:
                line_vertices = [_reproject(v[:2], self.crs, crs) for v in line]
                vertices.append(np.array(line_vertices))
            return vertices

    def get_coordinate_lists(self, crs=None):
        """ Returns a list of 2xn arrays representing lists of coordinates """
        ret = []
        for line_vertices in self.get_vertices(crs=crs):
            ret.append(line_vertices.T)
        return ret

class Multipolygon(Multipart, MultiVertexMultipartMixin, GeoJSONOutMixin,
                   ShapefileOutMixin):
    """ Collection of polygons with associated attributes.

    Parameters
    ----------
    coords : list
        list of lists of polygon rings, each consisting of 2-tuples or 3-tuples
    data : list, dict, Table object, or None
        point-specific data [default None]
    properties : dict or None
        geometry specific data [default None]
    crs : karta.crs.CRS subclass
        [default Cartesian]
    """
    def __init__(self, vertices, build_index=True, **kwargs):
        if len(vertices) == 0:
            self.vertices = []
        elif isinstance(vertices[0], Polygon):
            self.vertices = []
            for polygon in vertices:
                rings = [polygon.vertices]
                for sub in polygon.subs:
                    rings.append(sub)
                self.vertices.append(rings)
            kwargs.setdefault("crs", vertices[0].crs)
        else:
            self.vertices = []
            for part in vertices:
                rings = [CoordString(ring) for ring in part]
                self.vertices.append(rings)
        super(Multipolygon, self).__init__(vertices, **kwargs)
        if build_index:
            self.rtree = RTree([v[0] for v in self.vertices])
        self._geotype = "Multipolygon"
        return

    def __getitem__(self, key):
        if isinstance(key, numbers.Integral):
            properties = self.d[key]
            properties.update(self.properties)
            subs = []
            for vertices in self.vertices[key][1:]:
                subs.append(Polygon(vertices, properties=properties, crs=self.crs))
            vertices = self.vertices[key][0]
            return Polygon(vertices, subs=subs, properties=properties, crs=self.crs)
        elif isinstance(key, slice):
            return Multipolygon(self.vertices[key], properties=self.properties,
                                data=self.d[key], crs=self.crs)
        else:
            raise KeyError(type(key))

    @property
    def geomdict(self):
        return {"type" : "MultiPolygon",
                "bbox" : self.bbox,
                "coordinates" : _as_nested_lists(self.vertices)}

    @property
    def __geo_interface__(self):
        p = dict(_karta_proj4=self.crs.get_proj4())
        return {"type": "Feature",
                "geometry": self.geomdict,
                "properties": p}

    @property
    def vertices_ring(self):
        # inefficient implementation
        vertices = []
        for poly in self.vertices:
            rings = []
            for ring in poly:
                xys = [xy for xy in ring]
                xys.append(ring[0])
                rings.append(CoordString(xys))
            vertices.append(rings)
        return vertices

    def get_vertices(self, crs=None):
        """ Return vertices as a list of arrays.

        Parameters
        ----------
        crs : karta.CRS, optional
            coordinate system of output vertices
        """
        if crs is None or (crs == self.crs):
            vertices = []
            for poly_vertices in self.vertices:
                vertices.append([v.asarray() for v in poly_vertices])
            return vertices
        else:
            vertices = []
            for poly_vertices in self.vertices:
                poly = []
                for ring_vertices in poly_vertices:
                    poly.append(np.array([_reproject(v[:2], self.crs, crs)
                                          for v in ring_vertices]))
                vertices.append(poly)
            return vertices

    def get_coordinate_lists(self, crs=None):
        """ Returns a list of 2xn arrays representing lists of coordinates """
        ret = []
        for poly_vertices in self.get_vertices(crs=crs):
            poly = []
            for ring_vertices in poly_vertices:
                poly.append(ring_vertices.T)
            ret.append(poly)
        return ret

def _sign(a):
    """ Return the sign of *a* """
    if a == 0.0:
        return 1
    else:
        return a/abs(a)

def multipart_from_singleparts(parts, crs=None):
    """ Merge singlepart geometries into a multipart geometry.
    Properties contained by all inputs are stored in Multipart data attribute.

    Parameters
    ----------
    parts : iterable of singlepart Geometry instances
        e.g. list of Points, Lines, or Polygons
    crs : karta.CRS
        coordinate system of output Geometry

    Returns
    -------
    Multipart
        e.g. Multipoint, Multiline, or Multipolygon
    """
    if len(parts) == 0:
        raise ValueError("cannot construct multipart from zero singleparts")

    if crs is None:
        crs = parts[0].crs

    keys = list(parts[0].properties.keys())
    for part in parts[1:]:
        for key in keys:
            if key not in part.properties:
                keys.pop(keys.index(key))

    if len(keys) == 0:
        data = None
    else:
        data = {}
        for key in keys:
            data[key] = [part.properties[key] for part in parts]

    gt = parts[0]._geotype
    if gt == "Point":
        vertices = [part.get_vertex(crs=crs) for part in parts]
        return Multipoint(vertices, data=data, crs=crs)
    elif gt == "Line":
        vertices = [part.get_vertices(crs=crs) for part in parts]
        return Multiline(vertices, data=data, crs=crs)
    elif gt == "Polygon":
        vertices = []
        for part in parts:
            part_vertices = [part.get_vertices(crs=crs)]
            for sub in part.subs:
                part_vertices.append(sub.get_vertices(crs=crs))
            vertices.append(part_vertices)
        return Multipolygon(vertices, data=data, crs=crs)
    else:
        raise GeometryError("cannot convert type '{0}' to multipart".format(gt))

def merge_multiparts(*multiparts, **kw):

    """ Merge a list of multipart geometries of the same type and return the
    combined geometry. Shared data keys are retained. Properties are combined,
    with the first definition of a duplicate property persisting.

    Parameters
    ----------
    multiparts... : Multipart
        multipart instances to merge
    crs : karta.crs.CRS, optional
        CRS to use, if provided. By default, the CRS of the first multipart is used.

    Returns
    -------
    Multipart instance with the same type as the input *multiparts*

    Raises
    ------
    TypeError if inputs are not Multipart
    GeometryError if inputs are not all of the same derived type
    """
    if len(multiparts) == 1:
        if isinstance(multiparts[0], Multipart):
            return multiparts[0]
        else:
            raise TypeError("first parameter not instance of Multipart")

    t = type(multiparts[0])
    if not all(isinstance(mp, t) for mp in multiparts[1:]):
        raise GeometryError("not all inputs are the same kind of geometry")

    crs = kw.get("crs", None)
    if crs is None:
        crs = multiparts[0].crs

    vertices = multiparts[0].get_vertices(crs=crs)
    _p = multiparts[0].properties
    keys = set(multiparts[0].data.fields)
    for mp in multiparts[1:]:
        vertices = np.vstack([vertices, mp.get_vertices(crs=crs)])

        p = mp.properties
        p.update(_p)
        _p = p

        keys = keys.intersection(mp.data.fields)

    d = {}
    for k in keys:
        v = []
        for mp in multiparts:
            v.extend(mp.d[k])
        d[k] = v

    return t(vertices, properties=p, data=d, crs=crs)

def affine_matrix(mpa, mpb):
    """ Compute the affine transformation matrix that best matches two
    Multipoint geometries using a least squares fit.

    Output is relative to the coordinate system of the first geometry, if they
    differ.

    Parameters
    ----------
    mpa, mpb : Multipoint
        matching length collection of control points to match
    """
    if len(mpa) != len(mpb):
        raise GeometryError("Input geometries must have identical length")
    vecp = np.asarray(mpb.get_vertices(mpa.crs)).ravel()
    A = np.empty([2*len(mpa), 6], dtype=np.float64)
    for i, (x, y) in enumerate(mpa.get_vertices()):
        A[2*i:2*i+2,:] = np.kron(np.eye(2), [x, y, 1])
    M, res, rank, singvals = np.linalg.lstsq(A, vecp, rcond=None)
    return np.reshape(M, [2, 3])




"""
2D raster functions
"""

import numpy as np
from .grid import RegularGrid

def _slope(D, res=(1.0, 1.0)):
    """ Return the scalar slope at each pixel using the neighbourhood method.
    http://webhelp.esri.com/arcgisdesktop/9.2/index.cfm?TopicName=How%20Slope%20works
    """
    dx, dy = res
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0 * dx)
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0 * dy)
    return np.pad(np.sqrt(Ddx*Ddx + Ddy*Ddy), ((1, 1), (1, 1)), "reflect",
                  reflect_type="odd")

def slope(grid, band=0):
    """ Return the scalar slope at each pixel using the neighbourhood method.

    Parameters
    ----------
    grid : RegularGrid
    band : int, optional
        band to compute slope for (default 0)

    Returns
    -------
    RegularGrid

    Notes
    -----
    http://webhelp.esri.com/arcgisdesktop/9.2/index.cfm?TopicName=How%20Slope%20works
    """
    if grid.skew != (0, 0):
        raise NotImplementedError("slope calculations not implemented on skewed grids")
    D = np.where(grid.data_mask, grid[:,:,band].astype(np.float32), np.nan)
    return RegularGrid(grid.transform, _slope(D, grid.resolution),
                       crs=grid.crs, nodata_value=np.nan)

def _aspect(D, res=(1.0, 1.0)):
    """ Return the slope aspect for each pixel.
    http://webhelp.esri.com/arcgisdesktop/9.2/index.cfm?TopicName=How%20Aspect%20works
    """
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0 * res[0])
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0 * res[1])
    return np.pad(np.arctan2(Ddy, -Ddx), ((1, 1), (1, 1)), "constant",
                  constant_values=(np.nan,))

def aspect(grid, band=0):
    """ Compute grid aspect.

    Parameters
    ----------
    grid : RegularGrid
    band : int, optional
        band to compute aspect for (default 0)

    Returns
    -------
    RegularGrid
    """
    if grid.skew != (0, 0):
        raise NotImplementedError("aspect calculations not implemented on skewed grids")
    D = np.where(grid.data_mask, grid[:,:,band].astype(np.float32), np.nan)
    return RegularGrid(grid.transform, _aspect(D, grid.resolution),
                       crs=grid.crs, nodata_value=np.nan)

def _grad(D, res=(1.0, 1.0)):
    """ Computes the gradient of potential D. Return a tuple (dx, dy).
    """
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0 * res[0])
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0 * res[1])
    return (np.pad(Ddx, ((1, 1), (1, 1)), "constant", constant_values=(np.nan,)),
            np.pad(Ddy, ((1, 1), (1, 1)), "constant", constant_values=(np.nan,)))

def gradient(grid, band=0):
    """ Compute gradient field from a grid.

    Parameters
    ----------
    grid : RegularGrid
    band : int, optional
        (default 0)

    Returns
    -------
    (RegularGrid, RegularGrid)
    """
    if grid.skew != (0, 0):
        raise NotImplementedError("gradient calculations not implemented on skewed grids")
    D = np.where(grid.data_mask, grid[:,:,band].astype(np.float32), np.nan)
    dDdx, dDdy = _grad(D, grid.resolution)
    return (RegularGrid(grid.transform, dDdx, crs=grid.crs, nodata_value=np.nan),
            RegularGrid(grid.transform, dDdy, crs=grid.crs, nodata_value=np.nan))

def _div(U, V, res=(1.0, 1.0)):
    """ Calculate the divergence of a vector field. """
    dUdx = (U[:,2:] - U[:,:-2]) / (2.0*res[0])
    dVdy = (V[2:,:] - V[:-2,:]) / (2.0*res[1])
    divergence = np.pad(dUdx, ((0, 0), (1, 1)), "constant", constant_values=(np.nan,)) \
               + np.pad(dVdy, ((1, 1), (0, 0)), "constant", constant_values=(np.nan,))
    return divergence

def divergence(grid, bands=(0, 1)):
    """ Compute divergence from a grid.

    Parameters
    ----------
    grid : RegularGrid
    band : tuple of ints, optional
        indicates orthogonal velocity bands (default (0, 1))

    Returns
    -------
    RegularGrid
    """
    if grid.skew != (0, 0):
        raise NotImplementedError("divergence calculations not implemented on skewed grids")
    return RegularGrid(grid.transform,
                       _div(grid[:,:,bands[0]],
                            grid[:,:,bands[1]],
                            grid.resolution),
                       crs=grid.crs, nodata_value=np.nan)

def _normed_potential_vectors(D, res=(1.0, 1.0)):
    """ Computes a U,V vector field of potential D. Scalar components of
    U,V are normalized to ``max(|U, V|)``.
    """
    dx, dy = res
    Ddx = ((2 * D[1:-1,2:] + D[:-2,2:] + D[2:,2:]) -
           (2 * D[1:-1,:-2] + D[:-2,:-2] + D[2:,:-2])) / (8.0*dx)
    Ddy = ((2 * D[2:,1:-1] + D[2:,2:] + D[2:,:-2]) -
           (2 * D[:-2,1:-1] + D[:-2,:-2] + D[:-2,2:])) / (8.0*dy)
    M = np.sqrt(Ddx**2 + Ddy**2)
    U = Ddx / M[np.isnan(M)==False].max()
    V = Ddy / M[np.isnan(M)==False].max()
    return (np.pad(U, ((1, 1), (1, 1)), "constant", constant_values=(np.nan,)),
            np.pad(V, ((1, 1), (1, 1)), "constant", constant_values=(np.nan,)))

def normed_potential_vectors(grid, band=0):
    """ Computes a U,V vector field of a potential grid. Scalar components of
    U,V are normalized to ``max(|U, V|)``.

    Parameters
    ----------
    grid : RegularGrid
    band : int, optional (default 0)

    Returns
    -------
    (RegularGrid, RegularGrid)
    """
    if grid.skew != (0, 0):
        raise NotImplementedError("potential vector calculations not implemented on skewed grids")
    D = np.where(grid.data_mask, grid[:,:,band].astype(np.float32), np.nan)
    u, v = _normed_potential_vectors(D, res=grid.resolution)
    return (RegularGrid(grid.transform, u, crs=grid.crs, nodata_value=np.nan),
            RegularGrid(grid.transform, v, crs=grid.crs, nodata_value=np.nan))

def hillshade(grid, azimuth=330.0, elevation=60.0, band=0):
    """ Return a hill-shaded version of *grid*.

    Parameters
    ----------
    grid : RegularGrid
    azimuth : float, optional
        direction of light source (default 330.0)
    elevation : float, optional
        height of light source (default 60.0)
    band : int, optional
        band to compute hillshade for (default 0)

    Returns
    -------
    RegularGrid

    Notes
    -----
    Currently assumes orthogonal coordinates.
    """
    dxgrid, dygrid = gradient(grid, band=band)
    dx = dxgrid[:,:,0]
    dy = dygrid[:,:,0]
    res = grid.resolution
    u = np.array((np.full_like(dx, res[0]), np.zeros_like(dx), dx))
    v = np.array((np.zeros_like(dy), np.full_like(dy, res[1]), dy))
    w = np.cross(u, v, axisa=0, axisb=0)
    wunit = w / np.atleast_3d(np.sqrt(np.sum(w**2, axis=-1)))

    s = np.array((np.cos(azimuth*np.pi/180.0),
                  np.sin(azimuth*np.pi/180.0),
                  np.sin(elevation*np.pi/180.0)))
    smat = np.full((wunit.shape[0], wunit.shape[1], 3), s, dtype=np.float64)
    dprod = (wunit*smat).sum(axis=-1)

    q = np.percentile(dprod[~np.isnan(dprod)], [2, 98])
    np.clip(dprod, q[0], q[1], out=dprod)

    return RegularGrid(grid.transform, values=dprod, crs=grid.crs)

